'''
Model parameters
=================


'''


import pyomo.environ as po
import pyomo.core.base.sets as poset
from pyomo.core.base.set_types import Reals
import itertools
import numpy as np
import pandas as pd
import wrapt
from collections import namedtuple

from grimsel.auxiliary.aux_m_func import set_to_list
import grimsel.core.io as io
from grimsel import _get_logger

logger = _get_logger(__name__)

class ParameterAdder:
    '''
    Takes care of initializing, setting, and resetting a single parameter.

    Parameters
    ----------
    parameter_name : str
        used as model attribute, also assumed to be the
        column name in case ``value_col==False``
    parameter_index : tuple
        tuple of pyomo sets as parameter index
    source_dataframe : str or pandas.DataFrame
        input dataframe (or its attribute name)
        containing the parameter values
    value_col : str
        name of the ``source_dataframe`` parameter value column;
        optional---is set to the parameter name if no value is provided
    mutable : bool
        passed to the ``pyomo.environ.Param`` initialization
    default : numeric
        parameter default value; passed to the ``pyomo.environ.Param``
        initialization

    '''


    def __init__(self, m, par):


        self.m = m
        self.parameter_name = par.parameter_name
        self.source_dataframe = par.source_dataframe
        self.filt_cols = par.filt_cols
        self.filt_vals = par.filt_vals
        self.default = par.default
        self.param_kwargs = {'mutable': par.mutable,
                             'default': par.default,
                             'domain': par.domain}

        log_str = ('Assigning parameter '
                   '{par} ...'.format(par=self.parameter_name))
        logger.info(log_str)

        self.parameter_index = ((par.parameter_index,)
                           if not isinstance(par.parameter_index, tuple)
                           else par.parameter_index)

        self.value_col = (par.value_col
                          if par.value_col else self.parameter_name)

        self.has_monthly_factors = (
                self.m.df_parameter_month is not None and
                self.parameter_name
                in self.m.df_parameter_month.parameter.unique())

        self.index_cols = (par.index_cols
                           if par.index_cols else self._get_index_cols())

        self.df, self.flag_infeasible = self._get_param_data()

        if (not self.flag_infeasible
            and not self.m.check_valid_indices(self.parameter_index)):
            self.flag_infeasible = True

#        # check if column exists in table
#        if not self.flag_infeasible and self.value_col not in self.df.columns:
#            logger.warning(' failed (column doesn\'t exist).')
#            self.flag_infeasible = True


        if self.has_monthly_factors:
            df_mt, sets_new = self._apply_monthly_factors(self.df)

            self._df = df_mt
            self.index_cols = list(sets_new)

            # get set objects
            self.parameter_index = (self.m.mt, *self.parameter_index)

            # modify IO class attribute to get the output table indices right
            io.table_struct.DICT_COMP_IDX[self.parameter_name] = sets_new

    @property
    def df(self):
        '''
        ``df`` is the DataFrame holding the parameter data. It is implemented
        as a property so it can't be changed after initialization of the
        :class:`ParameterAdder` instance. This allows to restore the original
        values using :func:`ParameterAdder.init_update()`.

        '''
        return self._df

    @df.setter
    def df(self, val):

        if hasattr(self, 'df'):
            raise RuntimeError('Trying to set frozen attribute '
                               'ParameterAdder.df')
        else:
            self._df = val

    @wrapt.decorator
    def _if_is_feasible(f, self, args, kwargs):
        if not self.flag_infeasible:
            return f(*args, **kwargs)
        else:
            pass

    @_if_is_feasible
    def init_update(self, *args):
        '''

        Parameters
        ----------
        args
            The :func:`_get_data_dict` parameters ``(df, monthly_fact_col)``

        '''

        data = self._get_data_dict(*args)

        if not hasattr(self.m, self.parameter_name):
            # new parameter
            log_str = ' ok.'

            self.param_kwargs['initialize'] = data
            setattr(self.m, self.parameter_name,
                    po.Param(*self.parameter_index, **self.param_kwargs)
                    )
        else:
            # update parameter values
            log_str = ' parameter exists: updating.'
            for key, val in data.items():
                getattr(self.m, self.parameter_name)[key] = val

        logger.info(log_str)

    def _get_data_dict(self, df=False, monthly_fact_col=None):
        '''
        Returns a data dictionary for internal or external data.

        '''

        if isinstance(df, bool) and not df:
            # case no df input -> use internal data
            df = self.df
        else:
            # case external df cols don't have mt_id but index_cols do
            if (not 'mt_id' in df.columns) and ('mt_id' in self.index_cols):
                # apply monthly factors to external df
                df, _ = (self._apply_monthly_factors(df) if not monthly_fact_col
                         else self._apply_monthly_factors(df, monthly_fact_col))
            else:
                # use external df as is
                df = df

        return df.loc[-df[self.value_col].isna()].set_index(self.index_cols)[self.value_col].to_dict()

    def _get_param_data(self):
        '''
        Performs various checks

        Parameters
        ----------
        source_dataframe : pandas.Dataframe


        '''

        df = pd.DataFrame()
        flag_empty = False

        if type(self.source_dataframe) is str:

            if getattr(self.m, self.source_dataframe, None) is not None:
                df = getattr(self.m, self.source_dataframe).copy()
                if df is None:
                    logger.warning('... failed (source_dataframe is None).')
                    df = pd.DataFrame()
                    flag_empty = True
            else:
                logger.warning('... failed (source_dataframe does not exist).')
                df = pd.DataFrame()
                flag_empty = True

        elif type(self.source_dataframe) is pd.DataFrame:
            df = self.source_dataframe.copy()

        if (not flag_empty) and (self.filt_cols and self.filt_vals):
            try:
                df = df.set_index(self.filt_cols).loc[list(self.filt_vals)]
                df = df.reset_index()
            except Exception as e:
                logger.error(('Got error {e} when trying to filter DataFrame '
                              '{df} by values {val} of columns {cols}. This '
                              'might happen e.g. if a power plant in the pp '
                              'but not in the lin set is missing a value for '
                              'pp_eff in the plant_encar table'
                              ).format(e=e, val=self.filt_vals,
                                       cols=self.filt_cols, df=df))


        # check if column exists in table
        if not flag_empty:
            if self.value_col not in df.columns:
                if self.default is not None:
                    # make nan column which will be filled with default later
                    df[self.value_col] = np.nan
                    logger.warning(' column doesn\'t exist but default '
                                   'value provided...')
                else:
                    logger.warning(' failed (column doesn\'t exist).')
                    flag_empty = True

        if not flag_empty:
            df = df[self.index_cols + [self.value_col]]

            if self.default is not None:
                df.loc[df[self.value_col].isna(), self.value_col] = self.default

        return df, flag_empty

    def _get_index_cols(self):
        '''
        Translate the ``parameter_index`` set objects to column names.

        '''

        # dictionary sets -> column names
        dict_ind = {'sy': 'sy', 'sy_hydbc': 'sy', 'nd': 'nd_id', 'ca': 'ca_id',
                    'pr': 'pp_id', 'chp': 'pp_id',
                    'ror': 'pp_id', 'pp': 'pp_id', 'add': 'pp_id',
                    'fl': 'fl_id', 'fl_prof': 'fl_id',
                    'ppall': 'pp_id', 'hyrs': 'pp_id', 'wk': 'wk_id',
                    'mt': 'mt_id',
                    'ndcnn': ['nd_id', 'nd_2_id', 'ca_id'],
                    'st': 'pp_id', 'lin': 'pp_id',
                    'ndfl_prof': ['nd_id', 'fl_id'],
                    'ndcafl': ['nd_id', 'ca_id', 'fl_id'],
                    'sy_ndca': ['sy', 'nd_id', 'ca_id'],
                    'pricesll_pf': 'price_pf_id', 'pricebuy_pf': 'price_pf_id',
                    'dmnd_pf': 'dmnd_pf_id', 'tmsy': ['tm_id', 'sy'],
                    'supply_pf': 'supply_pf_id',
                    'sy_pr_ca': ['sy', 'pp_id', 'ca_id']}

        # get list of columns from sets
        index_cols = [dict_ind[pi.getname()]
                      if type(pi) in [poset.SimpleSet, poset.OrderedSimpleSet]
                      else 'pp_id' # Exception: Set unions are always pp
                      for pi in self.parameter_index]

        index_cols = [[c] if not type(c) == list else c for c in index_cols]
        index_cols = list(itertools.chain.from_iterable(index_cols))

        return index_cols


    def _expand_to_months(self, df0):
        '''
        Adds an additional month column to the input df.

        Parameters
        ----------
        df0 : pd.DataFrame
            Table to be expanded

        Returns
        -------
        pd.DataFrame
            table with additional month columnd

        '''

        list_mts = self.m.df_parameter_month.mt_id.unique().tolist()

        sets = [c for c in df0.columns if not c == 'value']
        old_index = df0[list(sets)].apply(tuple, axis=1).tolist()
        new_index = list(itertools.product(list_mts, old_index))
        new_index = [(cc[0],) + cc[1] for cc in new_index]

        df1 = pd.DataFrame(new_index)

        return df1



    def _get_monthly_factors(self):
        '''
        Returns monthly factors DataFrame for joining with the
        original parameter data.

        '''

        param = self.parameter_name

        param_mask = self.m.df_parameter_month.parameter == param
        dff = self.m.df_parameter_month.loc[param_mask].copy()
        name_cols = [c for c in dff.columns if '_name' in c]


        if len(dff[name_cols].drop_duplicates()) > 1:
            raise ValueError('apply_monthly_factors: Detected inconsistent '
                             + f'sets for parameter {param}. '
                             + 'Each parameter must have one set group only.')

        # rename set_id columns using the set_name values
        set_dict = dff[name_cols].iloc[0].T.to_dict()
        set_dict = {kk.replace('name', 'id'): vv
                    for kk, vv in set_dict.items()}
        dff = dff.rename(columns=set_dict)

        return dff


    def _apply_monthly_factors(self, df, val_col='mt_fact'):
        '''
        Adds a monthly index to existing parameters.

        For example, fuel prices are defined as yearly averages in the initial
        parameter definition (``df_plant_node_encar`` table). Here the
        parameter is re-initialized using the monthly factors from the
        ``ModelBase.df_parameter_month`` input table.

        Parameters
        ----------
        param : str
            name of the model parameter attribute the monthly factors are
            applied to

        Raises
        ------
        KeyError
            If the ``param`` is not included in the
            ``io.table_struct.DICT_COMP_IDX``
            table |rarr| index dictionary and therefore its index set
            definition cannot be inferred.
        ValueError
            If a parameter in the ``df_parameter_month`` is associated with
            more than one set group, e.g. ``vc_fl`` with both
            ``(nd_id, fl_id)`` and ``(fl_id,)``

        '''

        param = self.parameter_name

        logger.info('Applying monthly factors to parameter %s'%param)

        try:
            sets_io = io.table_struct.DICT_COMP_IDX[param]
        except:
            raise KeyError(('ModelBase._apply_monthly_factors: '
                              + 'Parameter {} '
                              + 'not included in the IO '
                              + 'parameter list.').format(param))

        sets = tuple([st for st in sets_io if not st == 'mt_id'])
        sets_new = ('mt_id',) + sets

        df0 = df
        df1 = self._expand_to_months(df0)
        df1.columns = sets_new + (self.value_col,)

        # join monthly factors
        df_fact = self._get_monthly_factors().set_index(list(sets_new))[val_col]

        df1 = df1.join(df_fact, on=sets_new)
        df1[val_col] = df1[val_col].fillna(1)

        # apply monthly factor
        df1[self.value_col
            ] *= df1[val_col]

        return df1[list(sets_new + (self.value_col,))], sets_new


# %%
class Parameters:
    r'''
    Mixin class for the :class:`grimsel.core.model_base.ModelBase` class
    containing all parameter definitions.

    '''


    def _get_df_supply(self):
        '''
        Returns ``df_profsupply_soy`` table with ``'pp_id', 'ca_id'`` columns
        if not empty, else empty table wih corresponding columns.
        '''

        if getattr(self, 'df_profsupply_soy', pd.DataFrame()).empty:
            return pd.DataFrame(columns=['sy', 'pp_id', 'ca_id', 'value'])
        else:
            return self.translate_pf_id(
                       self.df_profsupply_soy.rename(
                           columns={'supply_pf_id': 'pf_id'}))

    def _get_df_demand(self):
        '''
        Returns ``df_profdmnd_soy`` table with ``'nd_id', 'ca_id'`` columns
        if not empty, else empty table wih corresponding columns.
        '''

        if self.df_profdmnd_soy.empty:
            return pd.DataFrame(columns=['sy', 'nd_id', 'ca_id', 'value'])
        else:
            return self.translate_pf_id(
                       self.df_profdmnd_soy.rename(
                           columns={'dmnd_pf_id': 'pf_id'}))



    def add_parameters(self):
        '''
        Adds all parameters to the model.

        Generates :class:`ParameterAdder` instances for each of the parameters
        and calls their :meth:`ParameterAdder.init_update` method.

        '''

        if __name__ == '__main__':
            self = ml.m

        fields = ('parameter_name', 'parameter_index', 'source_dataframe',
                  'value_col', 'filt_cols', 'filt_vals', 'mutable', 'default',
                  'index_cols', 'domain')
        defaults = (None,) * 6 + (True, 0, None, Reals)
        Par = namedtuple('Par', fields)
        Par.__new__.__defaults__ = defaults


        list_par = (
        Par('dmnd', self.sy_ndca, self._get_df_demand(), 'value'),
        Par('supprof', self.sy_pr_ca, self._get_df_supply(), 'value'),
        Par('chpprof', (self.sy_ndca),
            'df_profchp_soy', 'value'),
        Par('inflowprof', (self.sy_hyrs_ca | self.sy_ror_ca),
            'df_profinflow_soy', 'value',
            index_cols=['sy', 'pp_id', 'ca_id']),

        Par('pricebuyprof', (self.sy, self.pricebuy_pf),
            'df_profpricebuy_soy', 'value'),
        Par('pricesllprof', (self.sy, self.pricesll_pf),
            'df_profpricesll_soy', 'value'),

        Par('min_erg_mt_out_share', self.hyrs, 'df_hydro'),
        Par('max_erg_mt_in_share', self.hyrs, 'df_hydro', default=1),
        Par('min_erg_share', self.hyrs, 'df_hydro'),

        Par('weight', self.tmsy, 'df_tm_soy', default=1),
        Par('grid_losses', (self.nd, self.ca), 'df_node_encar'),

        Par('cap_trme_leg', (self.mt, self.ndcnn), 'df_node_connect'),
        Par('cap_trmi_leg', (self.mt, self.ndcnn), 'df_node_connect'),

        Par('vc_ramp', (self.ppall, self.ca), 'df_plant_encar', None,
            ['pp_id', 'ca_id'], set_to_list(self.rp_ca, [None, None])),

        Par('pp_eff', (self.pp - self.lin, self.ca), 'df_plant_encar', None,
            ['pp_id'], self.pp - self.lin, default=1),
        Par('cf_max', (self.pp, self.ca), 'df_plant_encar', None, ['pp_id'],
            self.pp),

        Par('erg_chp', (self.chp, self.ca), 'df_plant_encar', None, ['pp_id'],
            self.chp),

        Par('erg_inp', self.ndcafl,
            'df_fuel_node_encar'),
        Par('vc_fl', (self.fl, self.nd),
            'df_fuel_node_encar', default=0),

        Par('factor_lin_0', (self.lin, self.ca),
            'df_plant_encar', None, ['pp_id'], self.lin, default=0),
        Par('factor_lin_1', (self.lin, self.ca),
            'df_plant_encar', None, ['pp_id'], self.lin, default=0),

        Par('price_co2', self.nd, 'df_def_node'),
        Par('nd_weight', self.nd, 'df_def_node', default=1),

        Par('co2_int', self.fl, 'df_def_fuel', default=0),

        Par('cap_pwr_leg', (self.ppall_ca),
            'df_plant_encar', None, ['pp_id'], self.ppall,
            index_cols=['pp_id', 'ca_id']),
        Par('vc_om', (self.ppall, self.ca),
            'df_plant_encar', None, ['pp_id'], self.ppall),
        Par('fc_om', (self.add, self.ca),
            'df_plant_encar', None, ['pp_id'], self.add, default=0),
        Par('fc_cp_ann', (self.add, self.ca),
            'df_plant_encar', None, ['pp_id'], self.add),

        Par('pwr_pot', (self.add, self.ca),
            'df_plant_encar', None, ['pp_id'], self.add, default=1e9),
        Par('cap_avlb', (self.pp, self.ca),
            'df_plant_encar', None, ['pp_id'], self.pp, default=1),

        Par('st_lss_hr', (self.st, self.ca),
            'df_plant_encar', None, ['pp_id'], self.st),
        Par('st_lss_rt', (self.st, self.ca),
            'df_plant_encar', None, ['pp_id'], self.st),
        Par('discharge_duration', (self.st | self.hyrs, self.ca),
            'df_plant_encar', None, ['pp_id'], self.st | self.hyrs),

        Par('hyd_erg_bc', (self.sy_hydbc, self.hyrs), 'df_plant_month'),
        )


        self.dict_par = {}
        for par in list_par:
            parameter = ParameterAdder(self, par)
            self.dict_par[par.parameter_name] = parameter
            parameter.init_update()


    def reset_all_parameters(self):
        '''
        Reset all parameters to their original values.

        This can be used prior to the model parameter variations to reset
        all of the input data.

        '''

        for name, par in self.dict_par.items():

            logger.info('Resetting parameter {}'.format(name))

            par.init_update()



# %%

    @staticmethod
    def _make_model_parameters_doc():

        doc_dict = {
        'Model structure': {
                    'weight': r':math:`{w_\mathrm{\tau, t}} : \forall \mathrm{(\tau, t) \in tmsy}`: Time slot weight, indexed by time map :math:`\tau` and time slot :math:`t`.',
                    },
        'Profiles': {
                    'dmnd': r':math:`\Phi_\mathrm{dmd, t, n,c} : \forall \mathrm{(t,n,c) \in sy\_ndca}`: Demand profile',
                    'chpprof': r':math:`\Phi_\mathrm{chp, t, n,c} : \forall \mathrm{(t,n,c) \in sy\_ndca}`: CHP profile',
                    'supprof': r':math:`\Phi_\mathrm{supply, t, p,c} : \forall \mathrm{(t,p,c) \in sy\_pp\_ca}`: Supply (VRE) profile',
                    'inflowprof': r':math:`\Phi_\mathrm{inflow, t, p,c} : \forall \mathrm{(t,p,c) \in sy\_hyrs\_ca \cup sy\_ror\_ca}`: Reservoir and run-of-river inflow profile',
                    'pricebuyprof': r':math:`\Phi_\mathrm{pbuy, t, \phi} : \forall \mathrm{(t,\phi) \in sy\times pf}`: Energy carrier price profile (buying)',
                    'pricesllprof': r':math:`\Phi_\mathrm{psell, t, \phi} : \forall \mathrm{(t,\phi) \in sy\times pf}`: Energy carrier price profile (selling)',
                    },
        'Hydro parameters': {
                    'min_erg_mt_out_share': r':math:`\rho_\mathrm{min\_erg\_out, p} : \forall \mathrm{p \in hyrs}`: Minimum monthly reservoir production as share of maximum monthly inflow :math:`\rho_\mathrm{max\_erg\_in, hyrs}` :math:`\mathrm{(-)}`.',
                    'max_erg_mt_in_share': r':math:`\rho_\mathrm{max\_erg\_in, p} : \forall \mathrm{p \in hyrs}`: Maximum monthly reservoir inflow :math:`\mathrm{(-)}`.',
                    'min_erg_share': r':math:`\rho_\mathrm{min\_cap,p} : \forall \mathrm{p \in hyrs}`: Maximum monthly reservoir inflow :math:`\mathrm{(-)}`.',
                    'hyd_erg_bc': r':math:`\rho_\mathrm{hyd\_bc, t,p} : \forall \mathrm{(t,p) \in sy\_hydbc \times hyrs}`: Hydro filling level boundary conditions for specific hours as share of energy capacity :math:`\mathrm{(-)}`.',
                    },
        'Grid and transmission': {
                    'cap_trme_leg': r':math:`P_\mathrm{exp, m,n_1,n_2,c} : \forall \mathrm{(m,n_1,n_2,c) \in mt \times ndcnn}`: Internodal monthly export capacity.',
                    'cap_trmi_leg': r':math:`P_\mathrm{imp, m,n_1,n_2,c} : \forall \mathrm{(m,n_1,n_2,c) \in mt \times ndcnn}`: Internodal monthly import capacity.',
                    'grid_losses': r':math:`\mathrm{\eta_{grid, n,c}} : \forall \mathrm{(n,c) \in nd\_ca}`: Grid losses for each node and energy carrier.',
        },
        'Technical properties of assets': {
                    'cap_pwr_leg': r':math:`P_\mathrm{leg, p,c}: \forall \mathrm{(p,c) \in ppall\_ca}`: Legacy power plant capacity :math:`\mathrm{(MW)}`.',
                    'discharge_duration': r':math:`\zeta_\mathrm{p,c}: \forall\mathrm{(p,c) \in st\_ca \cup hyrs\_ca}`: Ratio of energy capacity and power capacity for energy storing assets :math:`\mathrm{(hours)}`.',
                    'pp_eff': r':math:`\eta_\mathrm{p,c}: \forall \mathrm{(p,c)\in pp\_ca\setminus lin\_ca}`: Power plant efficiency for constant supply curves :math:`\mathrm{(MWh_{c}/MWh_{fl})}`.',
                    'erg_chp': r':math:`e_\mathrm{chp,p,c} : \forall \mathrm{(p,c) \in chp\_ca}`: Heat-driven electricity generation from co-generation plants :math:`\mathrm{(MWh_{el}/yr)}`; used to scale the CHP profile :math:`\Phi_\mathrm{chp,t,n,c}`.',
                    'erg_inp': r':math:`e_\mathrm{inp,n,c,f} : \forall \mathrm{(n,c,f) \in ndcafl}`: exogenous energy production by fuel and node :math:`\mathrm{(MWh_{el}/yr)}`.',
                    'st_lss_hr': r':math:`\mathrm{\epsilon_{hr, p,c}} : \forall \mathrm{(p,c) \in st\_ca}`: Hourly storage leakage losses :math:`\mathrm{(1/hr)}`.',
                    'st_lss_rt': r':math:`\mathrm{\epsilon_{rt, p,c} }: \forall \mathrm{(p,c) \in st\_ca}`: Storage round-trip losses :math:`\mathrm{(-)}`.',
                    'cap_avlb': r':math:`\mathrm{\alpha_{mt, p,c}}: \forall \mathrm{(p,c) \in pp\_ca}`: Relative monthly capacity availability of dispatchable plants.',
                    'pwr_pot': r':math:`P_\mathrm{pot, p,c}: \add \mathrm{(p,c) \in add\_ca}`: Upper bound capacity added :math:`\mathrm{(MW)}`.',
                    },
        'Specific costs': {
                    'vc_ramp': r':math:`\mathrm{vc_{ramp, p,c}} : \forall \mathrm{(p,c) \in ppall\_ca}`: Specific variable ramping cost :math:`\mathrm{(EUR/MW)}`.',
                    'vc_fl': r':math:`\mathrm{vc_{f,n}} : \forall \mathrm{(f,n) \in fl\_nd}`: Specific fuel cost in each node :math:`\mathrm{(EUR/MWh_{fl})}`.',
                    'vc_om': r':math:`\mathrm{vc_{om, p,c}} : \forall \mathrm{(p,c) \in ppall\_ca}`: Specific variable O\&M costs :math:`\mathrm{(EUR/MWh_{el})}`.',
                    'fc_om': r':math:`\mathrm{fc_{om, p,c}} : \forall \mathrm{(p,c) \in ppadd\_ca}`: Specific fixed O\&M costs :math:`\mathrm{(EUR/MW/yr)}`.',
                    'fc_cp_ann': r':math:`\mathrm{fc_{cp, p,c}} : \forall \mathrm{(p,c) \in ppadd\_ca}`: Annualized specific capital investment costs :math:`\mathrm{(EUR/MW/yr)}`.',
                    },
        'Linear supply curve coefficients': {
                    'factor_lin_0': r':math:`f_\mathrm{0, p,c} : \forall \mathrm{(p,c) \in lin\_ca}`: Zero-order linear supply curve efficiency coefficient :math:`\mathrm{(MWh_{fl}/MWh_{el})}`.',
                    'factor_lin_1': r':math:`f_\mathrm{1, p,c} : \forall \mathrm{(p,c) \in lin\_ca}`: First-order linear supply curve efficiency coefficient :math:`\mathrm{(MWh_{fl}/MWh_{el}/MW_{el})}`.',
                    },
        'Emission parameters': {
                    'price_co2': r':math:`\pi_\mathrm{CO_2, n} : \forall \mathrm{n \in nd}`: Node-specific |CO2| price :math:`\mathrm{(EUR/t_{CO_2})}`.',
                    'co2_int': r':math:`i_\mathrm{CO_2, f} : \forall \mathrm{f \in fl}`: Fuel-specific |CO2| intensity :math:`\mathrm{(t_{CO_2}/MWh_{fl})}`.',
                    }}

        doc_dict_origin = {'cap_avlb': 'merge(df_plant_encar, df_parameter_month)',
         'cap_pwr_leg': 'df_plant_encar',
         'cap_trme_leg': 'df_node_connect',
         'cap_trmi_leg': 'df_node_connect',
         'chpprof': 'df_profchp_soy',
         'co2_int': 'df_def_fuel',
         'price_co2': 'merge(df_def_node, df_parameter_month)',
         'discharge_duration': 'df_plant_encar',
         'dmnd': 'df_profdmnd_soy',
         'erg_chp': 'df_plant_encar',
         'erg_inp': 'df_fuel_node_encar',
         'factor_lin_0': 'df_plant_encar',
         'factor_lin_1': 'df_plant_encar',
         'fc_cp_ann': 'df_plant_encar',
         'fc_om': 'df_plant_encar',
         'grid_losses': 'df_node_encar',
         'hyd_erg_bc': 'df_plant_month',
         'inflowprof': 'df_profinflow_soy',
         'max_erg_mt_in_share': 'df_hydro',
         'min_erg_mt_out_share': 'df_hydro',
         'min_erg_share': 'df_hydro',
         'pp_eff': 'df_plant_encar',
         'pricebuyprof': 'df_profpricebuy_soy',
         'pricesllprof': 'df_profpricesll_soy',
         'st_lss_hr': 'df_plant_encar',
         'st_lss_rt': 'df_plant_encar',
         'supprof': 'df_profsupply_soy',
         'vc_fl': 'merge(df_fuel_node_encar, df_parameter_month)',
         'vc_om': 'df_plant_encar',
         'vc_ramp': 'df_plant_encar',
         'weight': 'df_tm_soy',
         'pwr_pot': 'df_plant_encar',
         }

        df = pd.DataFrame(doc_dict).stack().reset_index()
        df.columns = ['par', 'cat', 'doc']


        df['tab'] = df.par.replace(doc_dict_origin).apply(lambda x: '``%s``'%x)

        df.par = df.par.apply(lambda x: '``%s``'%x)


        cols=['Parameter', 'Symbol', 'Domain', 'Doc', 'Table']
        df = pd.DataFrame(
                list(df.apply(lambda x: (x['cat'], x.par,) + tuple(x.doc.split(': ')) + (x.tab,), axis=1).values),
                columns=['cat'] + cols).set_index('cat')

        df['Symbol'] = df.Symbol.apply(lambda x: r'{}'.format(x).strip(' ')) + '`'
        df['Domain'] = ':math:`' + df.Domain.apply(lambda x: r'{}'.format(x))


        import tabulate

        doc_str = ''
        for cat in doc_dict.keys():
            table_str = tabulate.tabulate(df.loc[df.index.get_level_values('cat') == cat],
                                     tablefmt='rst', showindex=False,
                                     headers=cols)
            table_str = table_str.replace('\n', '\n    ')

            table_dir_str = '.. table:: **{}**\n\n    '.format(cat)
            table_str = table_dir_str + table_str

            doc_str += ('\n'*2)

            doc_str += table_str

        return doc_str


Parameters.__doc__ += '\n'*4 + Parameters._make_model_parameters_doc()





# -*- coding: utf-8 -*-
"""
Created on Fri Jul  6 11:23:06 2018

@author: martin-c-s
"""


class ModelLoopModifier():
    '''
    The purpose of this class is to modify the parameters of the BaseModel
    objects in dependence on the current run_id.

    This might include the modification of parameters like CO2 prices,
    capacities, but also more profound modifications like the
    redefinition of constraints.
    '''

    def __init__(self, ml):
        '''
        To initialize the ModelBase object is made an instance attribute.
        '''

        self.ml = ml

    def set_value_co2_price(self, dict_co2=None):
        '''
        Example method changing the CO2 prices.
        '''


        sw_name = 'swco'

        if dict_co2 is None:
            dict_co2 = {0: 40, 1: 5, 2: 80}

        slct_co2 = dict_co2[self.ml.dct_step[sw_name]]

        for kk in self.ml.m.price_co2:
            self.ml.m.price_co2[kk].value = slct_co2

        self.ml.dct_vl[sw_name + '_vl'] = str(slct_co2) + 'EUR/t_CO2'




'''
The timemap module contains the TimeMap class for

* mapping between hours of the year and model time slots of the year,
* convenient mapping between time slots/hours and other temporal indices
  (month ids, week ids, seasons, hours of the week/month, etc)

'''

import pandas as pd
import numpy as np
from grimsel import _get_logger

logger = _get_logger(__name__)

MONTH_DICT = {0:'JAN', 1:'FEB', 2:'MAR', 3:'APR', 4:'MAY', 5:'JUN',
               6:'JUL', 7:'AUG', 8:'SEP', 9:'OCT', 10:'NOV', 11:'DEC'}

SEASON_DICT = {'JAN': 'WINTER', 'FEB': 'WINTER', 'DEC': 'WINTER',
               'MAR': 'SPRING', 'APR': 'SPRING', 'MAY': 'SPRING',
               'JUN': 'SUMMER', 'JUL': 'SUMMER', 'AUG': 'SUMMER',
               'SEP': 'FALL', 'OCT': 'FALL', 'NOV': 'FALL'}
DOW_DICT = {0: 'MON', 1: 'TUE', 2: 'WED', 3: 'THU',
            4: 'FRI', 5: 'SAT', 6: 'SUN'}
DOW_TYPE_DICT = {**{d: 'WEEKDAY' for d in range(5)}, **{5: 'SAT', 6: 'SUN'}}

TM_DICT = {}

def _tm_hash(nhours, freq, start, stop, tm_filt):

    hash_val = hash((nhours, freq, start, stop, str(tm_filt)))
    return hash_val

class _UniqueInstancesMeta(type):
    '''
    Load ``TimeMap`` instance from the module dictionary, if exists.

    To avoid the generation of multiple time maps with the same
    parameters, they are stored in the ``TM_DICT`` module attribute
    and retrieved when appropriate.

    '''

    def __call__(cls, nhours=1, freq=1,
                 start='2015-1-1 00:00', stop='2015-12-31 23:59',
                 tm_filt=False, *args, **kwargs):

        key = _tm_hash(nhours, freq, start, stop, tm_filt)

        if key in TM_DICT:
            logger.warning(('TimeMap (%s) exists. Reading time '
                         'map from module dict.')%key)
            return TM_DICT[key]
        else:
            return super().__call__(nhours, freq, start, stop, tm_filt,
                                    *args, **kwargs)

class TimeMap(metaclass=_UniqueInstancesMeta):
    '''
    The ``TimeMap`` class generates various DataFrames for time mapping.

    For example, if we start from input data with 15 minutes (0.25 hour)
    time resolution and run the model with 2 hour time resolution, we
    need a ``TimeMap`` instance with ``freq=0.25`` and ``nhours=2``.

    After calling the :func:`gen_soy_timemap` (which occurs automatically
    in the ``__init__`` method) three attribute ``DataFrames`` are generated:

    * ``df_time_map``, indexed by the hour of the year *hy*
    * ``df_time_red``, reduced time map indexed by the time slot of the year *sy*
    * ``df_hoy_soy``, a minimal map between the *sy* and the *hy*

    Args
    ----
    nhours : float
        target time resolution in hours
    freq : float
        frequency of the base time map in hours
    start : str
        start datetime of time map; format ``2015-1-1 00:00``
    stop : str
        end datetime of time map; format ``2015-1-1 00:00``
    tm_filt : list
        list of tuples with shape ``(timemap_column, [values])``
        for time map filtering. For example,
        ``tm_filt=[('dow', [0, 1]), ('mt_id', [2])]`` generates a
        ``TimeMap`` instance whose tables only contain the days of the
        week 0 and 1 for the *mt_id* 2
    keep_datetime : bool
        Remove the *DateTime* column if False
    minimum : bool
        don't generate non-essential columns

    Raises
    ------
    AssertionError
        If ``nhours`` is not a multiple of ``freq``.

    Examples
    --------

    Initialize a ``TimeMap`` instance mapping the original 0.25 hour time
    resolution to 1.25 hours:

    >>> tm = TimeMap(nhours=1.25, freq=0.25, tm_filt=[('mt_id', [0]), ('hom', range(2))], minimum=True)

    Base ``df_time_map`` table:

    >>> print(tabulate.tabulate(tm.df_time_map, headers=tm.df_time_map.columns, tablefmt='rst'))
    ====  ======  =====  =======  =====  ======  =====  =====  =====  =======  ===========  ====  ====
      ..    hour    doy    mt_id    day    year    dow    how    hom    wk_id    wk_weight    hy    sy
    ====  ======  =====  =======  =====  ======  =====  =====  =====  =======  ===========  ====  ====
       0       0      1        0      1    2015      3     72      0        0          384  0        0
       1       0      1        0      1    2015      3     72      0        0          384  0.25     0
       2       0      1        0      1    2015      3     72      0        0          384  0.5      0
       3       0      1        0      1    2015      3     72      0        0          384  0.75     0
       4       1      1        0      1    2015      3     73      1        0          384  1        0
       5       1      1        0      1    2015      3     73      1        0          384  1.25     1
       6       1      1        0      1    2015      3     73      1        0          384  1.5      1
       7       1      1        0      1    2015      3     73      1        0          384  1.75     1
    ====  ======  =====  =======  =====  ======  =====  =====  =====  =======  ===========  ====  ====

    Reduced ``df_time_red`` table:

    >>> print(tabulate.tabulate(tm.df_time_red, headers=tm.df_time_red.columns, tablefmt='rst'))
    ====  ======  ====  =====  =====  =====  =====  ======  =====  ====  =======  ========  =======  ===========
      ..    year    sy    day    dow    doy    hom    hour    how    hy    mt_id    weight    wk_id    wk_weight
    ====  ======  ====  =====  =====  =====  =====  ======  =====  ====  =======  ========  =======  ===========
       0    2015     0      1      3      1      0       0     72  0           0      1.25        0          384
       1    2015     1      1      3      1      1       1     73  1.25        0      0.75        0          384
    ====  ======  ====  =====  =====  =====  =====  ======  =====  ====  =======  ========  =======  ===========

    Minimal convenience table ``df_hoy_soy`` only containing the *hy* |rarr| *sy*
    map:

    >>> print(tabulate.tabulate(tm.df_hoy_soy, headers=tm.df_hoy_soy.columns, tablefmt='rst'))
    ====  ====  ====
      ..    sy    hy
    ====  ====  ====
       0     0  0
       1     0  0.25
       2     0  0.5
       3     0  0.75
       4     0  1
       5     1  1.25
       6     1  1.5
       7     1  1.75
    ====  ====  ====
    '''



    def __repr__(self):

        return 'TimeMap (%s)'%hash(self)

    def __hash__(self):

        return _tm_hash(self.nhours, self.freq, self.start,
                       self.stop, self.tm_filt)

    def __init__(self, nhours=1, freq=1,
                 start='2015-1-1 00:00', stop='2015-12-31 23:59',
                 tm_filt=False, keep_datetime=False, minimum=False):

        self.freq = freq
        self.num_freq = (float(self.freq[:-1])
                         if isinstance(self.freq, str) else self.freq)
        self.str_freq = '%sH'%self.num_freq

        self.nhours = nhours

        self.start, self.stop = start, stop

        self.tm_filt = tm_filt
        self.keep_datetime = keep_datetime

        self.minimum = minimum

        self.df_time_red = pd.DataFrame()
        self.df_hoy_soy = pd.DataFrame()
        self.df_time_map = pd.DataFrame()

        TM_DICT[hash(self)] = self

        if nhours:
            self.gen_soy_timemap()

    def gen_hoy_timemap(self):


        logger.info('Generating time map with freq='
                    '{} nhours={} from {} to {}'.format(self.freq, self.nhours,
                                                        self.start, self.stop))

        df_time_map = pd.DataFrame(index=pd.date_range(self.start, self.stop,
                                                       freq=self.str_freq))
        df_time_map = (df_time_map.reset_index()
                                  .rename(columns={'index': 'DateTime'}))

        df_time_map['hour'] = df_time_map['DateTime'].dt.hour
        df_time_map['doy'] = df_time_map['DateTime'].dt.dayofyear
        df_time_map['mt_id'] = df_time_map['DateTime'].dt.month - 1
        df_time_map['day'] = df_time_map['DateTime'].dt.day
        df_time_map['year'] = df_time_map['DateTime'].dt.year
        df_time_map['dow'] = df_time_map['DateTime'].dt.weekday
        df_time_map['how'] = df_time_map['dow'] * 24 + df_time_map['hour']
        df_time_map['hom'] = (df_time_map['day'] - 1) * 24 + df_time_map['hour']
        df_time_map['wk_id'] = df_time_map['DateTime'].dt.week - 1

        # add number of hours per week
        df_time_map = (df_time_map
            .join(df_time_map.pivot_table(values='how', index='wk_id', aggfunc=len)
            .rename(columns={'how': 'wk_weight'}), on='wk_id'))


        # remove February 29
        mask_feb29 = ((df_time_map.mt_id == 1) & (df_time_map.day == 29))
        df_time_map = df_time_map.loc[-mask_feb29].reset_index(drop=True)


        # add hour of the year column
        get_hoy = lambda x: ((x.hour + 24 * (x.doy - 1)
                             + x.DateTime.dt.minute / 60).rename('hy')
                             .reset_index())
        df_time_map['hy'] = (df_time_map.groupby(['year']).apply(get_hoy)
                                        .reset_index(drop=True)['hy'])
        if not self.minimum:

            df_time_map['month'] = df_time_map['DateTime'].dt.month
            df_time_map['mt'] = df_time_map['mt_id'].map(MONTH_DICT)
            df_time_map['season'] = df_time_map['mt'].map(SEASON_DICT)
            df_time_map['wk'] = df_time_map['wk_id']
            df_time_map['dow_name'] = df_time_map['dow'].replace(DOW_DICT)
            df_time_map['dow_type'] = df_time_map['dow'].replace(DOW_TYPE_DICT)

            # week of the month
            dfwom = df_time_map[['wk_id', 'mt_id']]
            dfwkmt = (df_time_map[['wk_id', 'mt_id']]
                            .pivot_table(index='wk_id', values=['mt_id'],
                                         aggfunc=np.median))
            dfwkmt = dfwkmt.rename(columns={'mt_id': 'wk_mt'})
            dfwom = (dfwom.drop('mt_id', axis=1).drop_duplicates()
                          .join(dfwkmt, on=dfwkmt.index.names))
            dfwk_max = (dfwom.pivot_table(index='wk_mt', values=['wk_id'],
                                         aggfunc=max)
                             .rename(columns={'wk_id': 'wk_max'})
                             .reset_index() + 1)
            dfwk_max = dfwk_max.set_index('wk_mt')
            dfwom = dfwom.join(dfwk_max, on=dfwk_max.index.names).fillna(0)
            dfwom['wom'] = dfwom['wk_id'] - dfwom['wk_max']
            dfwom = dfwom.set_index('wk_id')['wom']
            df_time_map = df_time_map.join(dfwom, on=dfwom.index.names)

            df_time_map_ndays = pd.DataFrame(df_time_map.loc[:,['mt', 'year', 'day']]
                                            .drop_duplicates()
                                            .pivot_table(values='day',
                                                         index=['year','mt'],
                                                               aggfunc=len))['day'].rename('ndays')
            df_time_map = df_time_map.join(df_time_map_ndays,
                                           on=df_time_map_ndays.index.names)


        # apply filtering
        mask = df_time_map.mt_id.apply(lambda x: True).rename('mask')
        if self.tm_filt:
            for ifilt in self.tm_filt:
                mask &= df_time_map[ifilt[0]].isin(ifilt[1])

        if mask.sum() == 0:
            raise RuntimeError('Trying to generate and TimeMap which is empty '
                               'after filtering.')

        self.df_time_map = df_time_map.loc[mask].reset_index(drop=True)


        if not self.keep_datetime:
            self.df_time_map = self.df_time_map.drop('DateTime', axis=1)


    def gen_soy_timemap(self):
        '''
        Reduces the original timemap to a lower time resolution.

        * Adds additional columns *weight* and *sy* to the ``df_time_map`` DataFrame:
            - *sy* are the time slot, each of which covers ``nhours/num_freq`` of the
              original *hy*-indexed ``df_time_map`` rows
            - The *weight* is the number of hours per reduced time slot. It is used
              to calculate energy from average power.
        * Generates an attribute ``df_time_red``, which is *sy* indexed. For any given
          *sy*, the minimum value of the temporal columns (e.g. *mt_id*, *how*, etc)
          over the corresponding *hy* is selected.

        '''

        assert (self.nhours / self.num_freq)%1 == 0, \
                ('TimeMap.gen_soy_timemap: The time slot duration nhours must '
                 'be a multiple of the original time map frequency freq. '
                 'num_freq=%f, nhours=%f'%(self.num_freq, self.nhours))

        if self.df_time_map.empty:
            self.gen_hoy_timemap()

        df_time_map = self.df_time_map

        # add soy column to dataframe, based on nhours
        len_rep = self.nhours / self.num_freq
        len_rge = np.ceil(len(df_time_map)/ len_rep)
        df_tm = pd.DataFrame(np.repeat(np.arange(len_rge), [len_rep]),
                             columns=['sy']).iloc[:len(df_time_map)]
        df_time_map['sy'] = df_tm

        # add weight column to dataframe
        df_weight = df_time_map.pivot_table(values=['hy'], index='sy',
                                            aggfunc=len)
        df_weight = df_weight.rename(columns={'hy': 'weight'}) * self.num_freq
        df_time_map = df_time_map.join(df_weight, on='sy')

        self.df_hoy_soy = df_time_map[['sy', 'hy']]

        if self.nhours == self.num_freq:
            self.df_time_red = df_time_map
        else:

            df_time_map_num = df_time_map.select_dtypes(include=['integer',
                                                                 'floating'])
            col_nonnum = [c for c in df_time_map.columns
                          if not c in df_time_map_num.columns]
            df_time_map_oth = df_time_map[col_nonnum + ['hy']].set_index('hy')

            self.df_time_red = (df_time_map_num
                        .pivot_table(aggfunc=min, index=['year', 'sy'])
                        .reset_index())

            self.df_time_red = self.df_time_red.join(df_time_map_oth, on='hy')

    def get_year_share(self):
        '''
        For a given time map, returns the share after applying ``tm_filt``.

        The ``tm_filt`` parameter of the :class:`TimeMap` class allows
        to limit the time map definition to certain months, days, etc.
        The year share corresponds to the duration of the filtered time map
        as a share of the total year length.

        Returns
        -------
        float : year share of the time map

        '''

        return len(self.df_time_red) / (8760 / self.nhours)

    def _get_dst_days(self, list_months=['MAR', 'OCT']):

        # filter by relevant months
        _df = self.df_time_map.loc[self.df_time_map.mt.isin(list_months)]
        mask_dy = _df.dow_name == 'SUN'

        _df = _df.loc[mask_dy].pivot_table(index=['mt', 'year'],
                                           values='doy', aggfunc=np.max)
        _df = _df.reset_index()[['year', 'mt', 'doy']]

        dict_dst = _df.set_index(['year', 'mt'])['doy'].to_dict()


        return dict_dst
# %%

if __name__ == '__main__':
    import tabulate
    import doctest

    doctest.testmod()



import os
import json
import pandas as pd
import numpy as np
import statsmodels.api as sm
import wrapt

@wrapt.decorator
def silence_pd_warning(f, self, args, kwargs):

    def f_new():
        pd.options.mode.chained_assignment = None
        ret = f(*args, **kwargs)
        pd.options.mode.chained_assignment = 'warn'
        return ret

    return f_new()


def print_full(x):
    with pd.set_option('display.max_rows', len(x)) as _:
        print(x)



def get_ols(df, add_constant=True, verbose=False):
    '''
    Get a simple linear regression, based on a two-columns dataframe.

    Parameters
    ----------
        add_constant: bool

    '''


    y = df.iloc[:, 1]
    X = df.iloc[:, 0]
    if add_constant:
        X = sm.add_constant(X) ## let's add an intercept (beta_0) to our model


    logit_model = sm.OLS(y,X.astype(float))


    result=logit_model.fit()
    if verbose:
        print(result.summary())

    dfres = pd.concat([pd.concat([result.params], keys=['val'], names=['result']),
                       pd.concat([result.pvalues], keys=['p_val'], names=['result']),
                       pd.concat([pd.DataFrame([result])], keys=['result_object'], names=['result'])
                      ]).T

    return dfres

def expand_rows(df, all_cols, val_cols, asindex=False):
#    df = df_ren_spec_energy_slct_0
#    all_cols = ['bdg_type','bdg_period','scenario']
#    val_cols = [bdg_type,bdg_period,scenario]

    # for each column all_cols containing the key string 'all', copy the whole dataframe for each of the values in val_cols
    all_cols_slct = [x for x in all_cols if x in df.columns.tolist()]
    val_cols_slct = [val_cols[x] for x in range(len(val_cols)) if all_cols[x] in df.columns.tolist()]

    icol = 0
    df_out = df.copy()
    icol = 3
    for icol in range(len(all_cols_slct)):

        df_rest = df_out[df_out[all_cols_slct[icol]].apply(str) != 'all']
        df_expd = df_out[df_out[all_cols_slct[icol]].apply(str) == 'all']

        try:
            all_col_unq = df_expd[all_cols_slct[icol]].unique()
            if (all_col_unq[0] == 'all') and (len(all_col_unq) == 1):
                df_out = pd.DataFrame()

                ival = 1
                for ival in range(len(val_cols_slct[icol])):
                    df_add = df_expd.copy()
                    df_add[all_cols_slct[icol]] = val_cols_slct[icol][ival]

                    df_out= df_out.append(df_add)
                df_out = pd.concat([df_out.copy(),df_rest])
        except:
            print('Error expand_rows')


    if df_out.columns.size != df.columns.size:
        raise ValueError('expand_rows: Column number changed')
    else:
        pass

    if asindex:
        df_out = df_out.set_index(all_cols, append=True, drop=True)

    return df_out

def read_xlsx_table(wb, sheets, columns, value_col=None, sub_table='', drop=[]):
    ''' Read tables from Excel file '''

#    wb
#    sheets = ['PLANT_ENCAR']
#    columns=ppca_cols


    ncols = len(columns)
    table_complete = pd.DataFrame()
    sheetname = sheets[0]
    for sheetname in sheets:
        sheet = wb.sheet_by_name(sheetname)

        rows = []
        number_of_rows = sheet.nrows

        row = 7
        for row in range(number_of_rows):
            value0  = (sheet.cell(row,0).value)

            if value0 == '-->' + sub_table:
                row_add = [c.value if not c.ctype in [5] else np.nan for c in sheet.row(row)[1:ncols+1]]
                rows.append(row_add)

        table_complete = pd.concat([table_complete,
                                    pd.DataFrame(rows, columns=columns)])

        for icol in table_complete.columns:
            table_complete.loc[(table_complete[icol].apply(str) == 'Inf')
                             | (table_complete[icol].apply(str) == 'inf'),
                               [icol]] = float('inf')
    return table_complete[[c for c in table_complete if not c in drop]]


def translate_id(df, dft, n):
    '''
    translate id columns
    '''

    if type(n) == str:
        n = [n] * 2
    n_id = [i_n + '_id' for i_n in n]
    dft = dft[[n[0], n_id[0]]]
    dict_id = dft.set_index(n[0])[n[0] + '_id']
    df[n_id[1]] = df[n_id[1]].replace(dict_id)
    return df, dict_id





'''
Model sets
=================


'''


import pyomo.environ as po
import pyomo.core.base.sets as poset
import pandas as pd
import numpy as np

from grimsel.auxiliary.aux_general import silence_pd_warning
from grimsel.auxiliary.aux_m_func import cols2tuplelist
from grimsel import _get_logger

logger = _get_logger(__name__)


DICT_SETS_DOC = {r'sy': r'model time slots : df_tm_soy : t',
                 r'ppall': r'all power plant types : df_def_plant : p',
                 r'pp': r'dispatchable power plants with fuels : df_def_plant : p',
                 r'st': r'storage plants : df_def_plant : p',
                 r'pr': r'variable renewables with fixed profiles : df_def_plant : p',
                 r'ror': r'run-of-river plants : df_def_plant : p',
                 r'lin': r'dispatchable plants with linear supply curve : df_def_plant : p',
                 r'hyrs': r'hydro reservoirs : df_def_plant : p',
                 'chp': r'plants with co-generation : df_def_plant : p',
                 r'add': r'plants with capacity additions : df_def_plant : p',
                 r'rem': r'plants with capacity retirements : df_def_plant : p',
                 r'curt': r'dedicated curtailment technology : df_def_plant : p',
                 r'sll': r'plants selling produced energy carriers : df_def_plant : p',
                 r'rp': r'dispatchable plants with ramping costs : df_def_plant : p',
                 r'ppall_nd': r'combined :math:`\mathrm{ppall\times nd}` set; equivalent for all subsets of :math:`\mathrm{ppall}` : df_def_plant : (p,n)',
                 r'ppall_ndca': r'combined :math:`\mathrm{ppall\times nd\times ca}` set; equivalent for all subsets of :math:`\mathrm{ppall}` : merge(df_def_plant, df_plant_encar) : (p,n,c)',
                 r'ppall_ndcafl': r'combined :math:`\mathrm{ppall\times nd\times ca\times fl}` set; equivalent for all subsets of :math:`\mathrm{ppall}` : merge(df_def_plant, df_plant_encar) : (p,n,c,f)',
                 r'pp_ndcafl_sll': r'"fuels" sold by power plants :math:`\mathrm{pp}` consuming energy carrier :math:`\mathrm{ca}` : merge(df_def_plant, df_plant_encar) : (p,n,c,f)',
                 r'sy_hydbc\subset sy': r'Time slots with exogenously defined storage level boundary conditions. : df_plant_month : t',
                 r'mt': r'months : df_plant_month : m',
                 r'wk': r'weeks : df_plant_week : w',
                 r'ndcnn': r'combined node sets :math:`\mathrm{nd\times nd\times ca}` for inter-nodal transmission : df_node_connect : (n,n_2,c)',
                 r'symin_ndcnn': r'combined node sets :math:`\mathrm{sy\times nd\times nd\times ca}` for inter-nodal transmission : merge(df_tm_soy, df_node_connect) : (t,n,n_2,c)',
                 r'fl_erg': r'fuels with energy production constraints : df_def_fuel : f',
                 r'tm': r'time maps : df_tm_soy : \tau',
                 r'tmsy': r'combination of time maps and slots : df_tm_soy : (\tau,t)',
                 r'tmsy_mt': r'all relevant combinations :math:`\mathrm{tm\times sy\times mt}` : df_tm_soy : (\tau,t,m)',
                 r'sy_ppall_ca': r'combined :math:`\mathrm{sy\times ppall\times nd}` set; equivalent for all subsets of :math:`\mathrm{ppall}` : merge(df_plant_encar, df_tm_soy) : (t,p,c)',
                 r'nd': r'Nodes : df_def_node : n',
                 r'ca': r'Output energy carriers : df_def_encar : c',
                 r'fl': r'Fuels : df_def_fuel : f',
                 r'ndcafl': r'Relevant combinations of nodes, produced energy carriers, and fuels : df_node_fuel_encar : (n,c,f)',
                 r'pf': r'Profiles (demand, supply, price, etc) : df_def_profile : \phi',
                 r'sy_ndca': r'Combined :math:`\mathrm{sy\times nd\times ca}` set : merge(df_node_encar, df_tm_soy) : (t,n,c)',
                 r'pp_ndcaca': r'combined :math:`\mathrm{pp\times nd\times ca\times ca}` set describing plants which convert one produced energy carrier into another : merge(df_def_encar, df_def_plant, df_plant_encar) : (p,n,c_{out},c)'
                 }

class Sets:
    '''
    Mixin class for set definition.

    '''

    # base power plant subsets
    slct_sets = ['ppall', 'pp', 'st', 'pr', 'ror', 'lin',
                 'hyrs', 'chp', 'add', 'rem',
                 'curt', 'sll', 'rp']


    def define_sets(self):
        r'''
        Add all required sets to the model.

        Adds sets as defined by

        * the ``setlst`` dictionary initialized in the :func:`get_setlst`
          method
        * the DataFrame attributes of the :class:`ModelBase` class for more
          complex derived sets

        %s


        '''

        self.nd = po.Set(initialize=self.setlst['nd'])
        self.ca = po.Set(initialize=self.setlst['ca'])
        self.fl = po.Set(initialize=self.setlst['fl'])
        self.pf = po.Set(initialize=self.setlst['pf'])

        df_ndca = self.df_def_plant[['pp_id', 'nd_id']].set_index('pp_id')
        df_ndca = self.df_plant_encar[['pp_id', 'ca_id']].join(df_ndca,
                                                               on='pp_id')
        df_ndca = df_ndca[['pp_id', 'nd_id', 'ca_id']]

        slct_cols = ['pp_id', 'ca_id']


        for iset in self.slct_sets:

            logger.info('Defining basic sets for {}'.format(iset))

            ''' SUB SETS PP'''
            setattr(self, iset,
                    po.Set(within=(None if iset == 'ppall' else self.ppall),
                           initialize=self.setlst[iset])
                           if iset in self.setlst.keys()
                           else po.Set(within=self.ppall, initialize=[]))

            ''' SETS PP x ENCAR '''
            _df = self.df_plant_encar.copy()
            _df = _df.loc[_df['pp_id'].isin(getattr(self, iset))]
            setattr(self, iset + '_ca',
                    po.Set(within=getattr(self, iset) * self.ca,
                           initialize=cols2tuplelist(_df[slct_cols])))

            ''' SETS PP x ND x ENCAR '''
            _df = df_ndca.copy()
            _df = _df.loc[df_ndca['pp_id'].isin(self.setlst[iset]
                          if iset in self.setlst.keys() else [])]
            setattr(self, iset + '_ndca',
                    po.Set(within=getattr(self, iset) * self.nd * self.ca,
                           initialize=cols2tuplelist(_df)))


        # no scf fuels in the _cafl and _ndcafl
        # These are used to calculate fuel costs, that's why we don't want
        # the generated fuels in there.
        df_0 = self.df_def_plant[['pp_id', 'nd_id', 'fl_id']]
        df_0 = df_0.set_index('pp_id')
        df_0 = self.df_plant_encar[['pp_id', 'ca_id']].join(df_0, on='pp_id')
        df_0 = df_0.loc[df_0.fl_id.isin(self.setlst['fl'])]

        list_sets = ['ppall', 'hyrs', 'pp', 'chp', 'ror', 'st', 'lin']
#        list_sets = [st for st in list_sets if st in self.setlst.keys()]

        for iset in list_sets:

            if iset in self.setlst:
                cols_ppcafl = ['pp_id', 'ca_id', 'fl_id']
                df = df_0.loc[df_0['pp_id'].isin(self.setlst[iset]),
                              cols_ppcafl]
                new_set = po.Set(within=getattr(self, iset) * self.ca * self.fl,
                                 initialize=cols2tuplelist(df))
                setattr(self, iset + '_cafl', new_set)

                slct_cols_ppndcafl = ['pp_id', 'nd_id', 'ca_id', 'fl_id']
                df = df_0.loc[df_0.pp_id.isin(self.setlst[iset]),
                              slct_cols_ppndcafl]

                setattr(self, iset + '_ndcafl',
                        po.Set(within=(getattr(self, iset) * self.nd
                                     * self.ca * self.fl),
                               initialize=cols2tuplelist(df)))
            else:

                new_set_cafl = po.Set(within=getattr(self, iset) * self.ca * self.fl,
                                 initialize=[])
                setattr(self, iset + '_cafl', new_set_cafl)

                new_set_ndcafl = po.Set(within=getattr(self, iset) * self.nd * self.ca * self.fl,
                                        initialize=[])
                setattr(self, iset + '_ndcafl', new_set_ndcafl)


        # plants selling fuels ... only ppall, therefore outside the loop
        lst = cols2tuplelist(df.loc[df.pp_id.isin(self.setlst['sll']
                                                  if 'sll' in self.setlst
                                                  else [])])
        setattr(self, 'pp_ndcafl_sll',
                po.Set(within=self.pp_ndcafl, initialize=lst))

        # temporal
        self.sy = po.Set(initialize=list(self.df_tm_soy.sy.unique()),
                         ordered=True)

        self.sy_hydbc = (po.Set(within=self.sy,
                               initialize=set(self.df_plant_month.sy))
                         if not self.df_plant_month is None else None)

        self.mt = (po.Set(initialize=list(self.df_def_month['mt_id']))
                   if not self.df_def_month is None else None)
        self.wk = (po.Set(initialize=list(self.df_def_week['wk_id']))
                   if not self.df_def_week is None else None)

        # pp_cacafcl; used to account for conversions of ca in the supply rule
        if 'fl_id' in self.df_def_encar:
            df_cafl = self.df_def_encar.set_index('fl_id')['ca_id']
            df_cafl = df_cafl.rename('ca_fl_id')
            df_ppca = self.df_plant_encar.set_index('pp_id')['ca_id']
            df = (self.df_def_plant.join(df_ppca, on='pp_id')
                                   .join(df_cafl, on='fl_id'))
            df = df.loc[-df.ca_fl_id.isnull()
                      & -df.ca_id.isnull(), ['pp_id', 'nd_id', 'ca_id',
                                             'ca_fl_id']]
            self.pp_ndcaca = po.Set(within=self.pp_ndca * self.ca,
                                      initialize=cols2tuplelist(df))
        else:
            self.pp_ndcaca = None

        # inter-node connections
        if not self.df_node_connect is None and not self.df_node_connect.empty:
            df = self.df_node_connect[['nd_id', 'nd_2_id', 'ca_id']]
            self.ndcnn = po.Set(within=self.nd * self.nd * self.ca,
                             initialize=cols2tuplelist(df), ordered=True)

            df = self.df_symin_ndcnn[['symin', 'nd_id', 'nd_2_id', 'ca_id']]
            self.symin_ndcnn = po.Set(within=self.sy * self.nd
                                             * self.nd * self.ca,
                                      initialize=cols2tuplelist(df),
                                      ordered=True)
        else:
            self.ndcnn = po.Set(within=self.nd * self.nd * self.ca)
            self.symin_ndcnn = po.Set(within=self.sy * self.nd
                                             * self.nd * self.ca)


        # ndca for electricity only; mainly used for flexible demand;
        # then again: why would only EL have flexible demand?
        df = pd.concat([pd.Series(self.slct_node_id, name='nd_id'),
                        pd.Series(np.ones(len(self.slct_node_id))
                                  * self.mps.dict_ca_id['EL'],
                                  name='ca_id')], axis=1)
        self.ndca_EL = po.Set(within=self.nd * self.ca,
                              initialize=cols2tuplelist(df), ordered=True)

        # general ndca
        df = self.df_node_encar[['nd_id', 'ca_id']].drop_duplicates()
        self.ndca = po.Set(within=self.nd * self.ca,
                           initialize=cols2tuplelist(df), ordered=True)

        # general ndcafl
        if not self.df_fuel_node_encar is None:
            df = self.df_fuel_node_encar[['nd_id', 'ca_id', 'fl_id']]
            self.ndcafl = po.Set(within=self.nd * self.ca * self.fl,
                              initialize=cols2tuplelist(df), ordered=True)
        else:
            self.ndcafl = None

        # fuels with energy constraints
        if 'is_constrained' in self.df_def_fuel:
            lst = self.df_def_fuel.loc[self.df_def_fuel.is_constrained==1,
                                           'fl_id'].tolist()
            self.fl_erg = po.Set(within=self.fl, initialize=lst, ordered=True)
        else:
            self.fl_erg = po.Set(within=self.fl, initialize=[])

        # set pf_id for profiles
        for pf_set in ['dmnd_pf', 'supply_pf', 'pricesll_pf', 'pricebuy_pf']:
            setattr(self, pf_set,
                    po.Set(within=self.pf, initialize=self.setlst[pf_set],
                           ordered=True))

        self._init_tmsy_sets()


    def _init_tmsy_sets(self):
        '''

        The plant ids and the time slots are connected
        through the node-specific time resolution.
        '''

        self.tm = po.Set(initialize=self.df_tm_soy.tm_id.unique(),
                         ordered=True)

        list_tmsy = cols2tuplelist(self.df_tm_soy[['tm_id', 'sy']])
        self.tmsy = po.Set(within=self.tm*self.sy, initialize=list_tmsy,
                           ordered=True)

        # only constructed if self.mt exists
        self.tmsy_mt = (po.Set(within=self.tmsy * self.mt,
                               initialize=cols2tuplelist(
                                    self.df_tm_soy[['tm_id', 'sy', 'mt_id']]))
                        if not self.mt is None else None)

        df = pd.merge(self.df_def_node, self.df_node_encar,
                      on='nd_id', how='outer')[['nd_id', 'ca_id']]
        df = df.loc[~df.ca_id.isna()].drop_duplicates()
        df['tm_id'] = df.nd_id.replace(self.dict_nd_tm_id)
        cols = ['sy', 'nd_id', 'ca_id']
        list_syndca = pd.merge(self.df_tm_soy[['tm_id', 'sy']],
                                df, on='tm_id', how='outer')[cols]

        self.sy_ndca = po.Set(within=self.sy*self.ndca, ordered=True,
                              initialize=cols2tuplelist(list_syndca))

        mask_pp = self.df_plant_encar.pp_id.isin(self.setlst['ppall'])
        df = self.df_plant_encar.loc[mask_pp, ['pp_id', 'ca_id']].copy()
        df['tm_id'] = (df.pp_id
                         .replace(self.mps.dict_plant_2_node_id)
                         .replace(self.dict_nd_tm_id))
        cols = ['sy', 'pp_id', 'ca_id']
        list_syppca = pd.merge(self.df_tm_soy[['sy', 'tm_id']],
                                df, on='tm_id', how='outer')[cols]

        list_syppca = list_syppca.loc[~(list_syppca.pp_id.isna()
                                        | list_syppca.ca_id.isna())]

        list_syppca = cols2tuplelist(list_syppca)

        for slct_set in ['ppall', 'rp', 'st', 'hyrs', 'pr',
                         'pp', 'chp', 'ror', 'lin']:

            set_name = 'sy_%s_ca'%slct_set
            within = self.sy * getattr(self, slct_set) * self.ca

            if slct_set in self.setlst:

                logger.info('Defining set ' + set_name)

                set_pp = set(self.setlst[slct_set])

                setattr(self, set_name,
                        po.Set(within=within, ordered=True,
                               initialize=[row for row in list_syppca
                                           if row[1] in set_pp]))
            else:
                setattr(self, set_name, po.Set(within=within, initialize=[]))

    def get_setlst(self):
        '''
        Lists of indices for all model components are extracted from the
        input tables and stored in a dictionary ``ModelBase.setlst``.

        For the most part power plant subset definitions are based on the
        binary columns *set_def_..* in the ``df_def_plant`` input table.

        '''
        # define lists for set initialization
        self.setlst = {st: [] for st in self.slct_sets}

        df = self.df_def_plant

        qry = ' & '.join(['{} == 0'.format(sd)
                         for sd in ('set_def_tr', 'set_def_dmd')
                         if sd in df.columns])

        self.setlst['ppall'] = (df.query(qry).pp_id.tolist())

        for ippset in df.columns[df.columns.str.contains('set_def')]:
            # Note: index starting at 8 removes prefix set_def_ from col name
            self.setlst[ippset[8:]] = df.loc[df[ippset] == 1, 'pp_id'].tolist()

        mask_node = self.df_def_node['nd_id'].isin(self.slct_node_id)
        self.setlst['nd'] = self.df_def_node.loc[mask_node]['nd_id'].tolist()
        self.setlst['ca'] = self.df_def_encar.ca_id.tolist()

        # fuels are bought fuels only, not generated encars used as input
        df = self.df_def_fuel.copy()
        self.setlst['fl'] = df.fl_id.tolist()


        for col, df in [('supply_pf_id', self.df_plant_encar),
                        ('pricesll_pf_id', self.df_fuel_node_encar),
                        ('pricebuy_pf_id', self.df_fuel_node_encar),
                        ('dmnd_pf_id', self.df_node_encar)]:

            if col in df.columns:
                df_ = df.copy()
                df_ = df_.loc[-df_[col].isna(), col]
                self.setlst[col.replace('_id', '')] = df_.unique().tolist()
            else:
                self.setlst[col.replace('_id', '')] = []

        self.setlst['pf'] = (self.setlst['dmnd_pf']
                             + self.setlst['pricesll_pf']
                             + self.setlst['pricebuy_pf']
                             + self.setlst['supply_pf'])

        self.setlst['rp'] = ((self.setlst['pp'] if 'pp' in self.setlst else [])
                             + (self.setlst['ror'] if 'ror' in self.setlst else [])
                             + (self.setlst['hyrs'] if 'hyrs' in self.setlst else [])
                             + (self.setlst['st'] if 'st' in self.setlst else []))

    @silence_pd_warning
    @staticmethod
    def _get_set_docs():
        '''
        Convenience method to extract all set docs from a :class:`ModelBase`
        instance.

        '''

        import tabulate

        to_math = lambda x: ':math:`\mathrm{%s}`'%x

        comb_sets = ['ndcnn', 'tmsy']

        cols = ['Set', 'Members', 'Description', 'Source table']
        df_doc = pd.Series(DICT_SETS_DOC).reset_index()
        df_doc[['Description', 'Source', 'Members']] = pd.DataFrame(df_doc[0].apply(lambda x: tuple(x.split(' : '))).tolist())
        df_doc = df_doc.drop(0, axis=1)
        df_doc.columns = ['Set', 'Description', 'Source table', 'Members']
        df_doc.Members = df_doc.Members.apply(lambda x: ':math:`\mathrm{%s}`'%(x))
        df_doc['Source table'] = df_doc['Source table'].apply(lambda x: '``%s``'%(x))
        df_doc = df_doc[cols]


        mask_not_under = ~df_doc.Set.str.contains('_')
        mask_not_ndca = ~df_doc.Set.str.contains('_ndca')
        mask_not_ca = ~df_doc.Set.str.contains('_ca')


        list_pp = ['ppall', 'pp', 'st', 'pr', 'ror', 'lin', 'hyrs', 'chp',
                   'add', 'rem', 'curt', 'sll', 'rp']

        df_doc_oth = df_doc.loc[~df_doc.Set.isin(list_pp)
                                & ~df_doc.Set.isin(comb_sets)
                                & mask_not_under]
        df_doc_oth['Set'] = df_doc_oth.Set.apply(to_math)
        table_base = (tabulate.tabulate(df_doc_oth,
                                headers=cols,
                                tablefmt='rst', showindex=False))



        df_doc_pp = df_doc.loc[df_doc.Set.isin(list_pp)
                               & mask_not_under]
        df_doc_pp['Set'] = df_doc_pp.Set.apply(to_math)
        df_doc_pp.Set = df_doc_pp.Set.apply(lambda x: x.replace('}', '\subset ppall}').replace('ppall\subset ', ''))
        table_pp = (tabulate.tabulate(df_doc_pp,
                                tablefmt='rst', headers=cols,
                                showindex=False))



        df_doc_oth = df_doc.loc[(df_doc.Set.isin(comb_sets)
                                | ~mask_not_under)]
        df_doc_oth['Set'] = df_doc_oth.Set.apply(to_math)

        df_doc_ppca = df_doc.loc[~mask_not_ca | ~mask_not_ndca]
        df_doc_ppca['Set'] = df_doc_ppca.Set.apply(to_math)
        df_doc_ppca = df_doc_ppca.loc[df_doc_ppca.Set.str.contains('ppall')
                        | df_doc_ppca.Set.str.endswith('sll')]
        df_doc_oth = pd.concat([df_doc_ppca, df_doc_oth])
        df_doc_oth.Set = df_doc_oth.Set.apply(lambda x: x.replace('_', '\_'))
        table_derived = tabulate.tabulate(df_doc_oth, headers=cols,
                                tablefmt='rst', showindex=False)

        doc_str = r'''

.. table:: **Primary base sets**
    :widths: 15 10 100 30

    %s


.. _power_plant_sets:

.. table:: **Primary power plant sets and subsets**
    :widths: 15 10 100 30

    %s

.. table:: **Key derived sets**
    :widths: 15 10 100 30

    %s

.. note::

   * :math:`\mathrm{sy\_ppall\_ca}`: The operation (power production/charging)
     of a given plant :math:`\mathrm{pp}` is defined for each of the time
     slots of the
     corresponding node. Since the time resolution and hence the
     number of time slots potentially depends on the node, this
     combined set is necessary to limit the variable and constraint
     definitions to the relevant time slots of any power plant.
   * :math:`\mathrm{symin\_ndcnn}`: If two nodes with different time
     resolutions are connected, the transmission variable has the
     higher time resolution of the two (*min* as in "smallest time slot
     duration"). This combined set expresses this relationship for
     each of the connected nodes.

'''%(table_base.replace('\n', '\n    '),
           table_pp.replace('\n', '\n    '),
           table_derived.replace('\n', '\n    '))

        return doc_str




Sets.define_sets.__doc__ = Sets.define_sets.__doc__%Sets._get_set_docs()



'''
Model constraints
===================

All model constraints are defined in the :class:`Constraints` class, which
serves as a mixin class to :class:`grimsel.core.model_base.ModelBase`.

'''


from functools import wraps

import pyomo.environ as po

from grimsel.core.io import IO
from grimsel.auxiliary.aux_m_func import set_to_list
from grimsel import _get_logger

logger = _get_logger(__name__)



nnnn = [None] * 4
nnn = [None] * 3
nn = [None] * 2


def _limit_max_sy(dct):
    '''
    Decorator limiting the constraints to a timemap-dependent range.

    Note: Only useful if the sy_pp_ca sets are not defined (which they will
    probably always be, since they are required by the variables).

    Parameters
    ----------
    dct: dict
        maximum time slot in dependence on the other parameters
        (typically ``{nd: sy_max}`` or ``{pp: sy_max}``)

    '''
    def wrapper(f):
        @wraps(f)
        def wrapped(self, *args, **kwargs):
            if args[0] <= dct[args[1]]:
                return f(self, *args, **kwargs)
            else:
                return po.Constraint.Skip
        return wrapped
    return wrapper



class Constraints:
    '''
    Mixin class containing all constraints, included in the
    :class:`grimsel.core.model_base.ModelBase`
    '''

    def cadd(self, name, *args, objclass=po.Constraint, **kwargs):
        '''
        Add constraints or objectives to the model after logging.

        Parameters
        ----------
        name : str
            name of the new component
        objclass : pyomo class
            one of ``{po.Constraint, po.Objective}``
        args, kwargs
            passed to the ``objclass`` initialization

        '''

        ls = 'Adding {} {}: {}.'.format(objclass.__name__.lower(), name,
                                        kwargs['rule'].__doc__)
        logger.info(ls)

        obj = objclass(*args, **kwargs)
        setattr(self, name, obj)

    def add_transmission_bounds_rules(self):
        r'''
        Add transmission bounds.

        .. math::

           & -P_\mathrm{imp,m,n,n_2,c} \leqslant p_\mathrm{trm,t,n,n_2,c}
           \leqslant P_\mathrm{exp,m,n,n_2,c} \\
           & \forall \mathrm{(t,n,n_2,c) \in symin\_ndcnn} \\

        .. note::
           This method modifies the ``trm`` transmission power Pyomo
           variable object by calling its ``setub`` and ``setlb`` methods.

        '''

        dict_weight = IO.param_to_df(self.nd_weight).set_index('nd_id').value.to_dict()

        if hasattr(self, 'trm'):
            for sy, nd1, nd2, ca in self.trm:

                # get node with max weight
                nd_w = max([nd1, nd2], key=lambda x: dict_weight[x])

                tm = self.dict_ndnd_tm_id[nd1, nd2]

                mt = self.dict_soy_month[(tm, sy)]

                ub = (self.cap_trme_leg[mt, nd1, nd2, ca]
                      * self.nd_weight[nd_w])
                lb = - (self.cap_trmi_leg[mt, nd1, nd2, ca]
                        * self.nd_weight[nd_w])

                self.trm[(sy, nd1, nd2, ca)].setub(ub)
                self.trm[(sy, nd1, nd2, ca)].setlb(lb)

    def add_supply_rules(self):
        r'''
        Adds the supply rule: Balance supply/demand in each node for each
        time slot :math:`\mathrm{t}`, node :math:`\mathrm{n}` and,
        produced energy carrier :math:`\mathrm{c}`.

        * The supply side consists of the power production
          :math:`p_\mathrm{t,p,c}` from all plants producing energy
          carrier :math:`\mathrm{c}`. This is reduced by the curtailed
          and the sold power. Additionally, imports or exports from
          connected nodes enter this side of the equation depending on the
          directionality definition in the combined set :math:`\mathrm{ndcnn}`
          (see the **note** below).
        * The demand side consists of the exogenous demand profile, the storage
          charging power, the relevant imports and exports, depending on the
          direction definition, and the consumption of energy carrier
          :math:`\mathrm{c}` for the production of other energy carriers
          :math:`\mathrm{c_{out}}`.

        .. math::

           & \sum_\mathrm{ppall\setminus (sell\cup curt)} p_\mathrm{sy\_pp\_ca}
               - \sum_\mathrm{sell}p_\mathrm{sy\_sell\_ca}
               - \sum_\mathrm{curt}p_\mathrm{sy\_curt\_ca} \\
           & + \sum_\mathrm{nd_2} p_\mathrm{trm, nd_2 \rightarrow nd} \\
           & = \phi_\mathrm{dmd, t, n, c} \\
           & + \sum_\mathrm{p \in st_n} p_\mathrm{chg,t,p,c}\\
           & + \sum_\mathrm{nd_2} p_\mathrm{trm, nd \rightarrow nd_2} \\
           & + \sum_\mathrm{(p, n, c_{out}, c)\in pp\_ndcaca}
               p_\mathrm{t,p,c_{out}} / \eta_\mathrm{p,c_{out}} \\
           & \forall \mathrm{(t,n,c) \in sy\_nd\_ca} \\

        .. note::

            * **Directionality of inter-nodal transmission**: Transmission
              between nodes is expressed through the variables
              :math:`p_\mathrm{trm, symin_ndcnn}`. These variables can be
              positive or negative depending on the power flow direction.
              Since only one direction is included in the
              :math:`\mathrm{ndcnn}` set, e.g.

              .. math::

                 & \mathrm{(nd_1=0, nd_2=1, ca=0) \in ndcnn} \\
                 & \Rightarrow \mathrm{(nd_1=1, nd_2=0, ca=0) \notin ndcnn}, \\

              they enter the supply constraint on both sides.
            * **Transmission between nodes with different time
              resolutions**: The transmission power variable
              :math:`p_\mathrm{trm}` has the higher time resolution of the two
              connected nodes. On the side of the lower time resolution node,
              it is averaged over the corresponding time slots.
            * **Consumption of produced energy carriers** :math:`\mathrm{ca}`:
              Generators which consume an endogenously produced energy carrier
              (e.g. electricity) to produce another energy carrier (e.g. heat)
              enter the supply constraint on the demand side
              as :math:`\sum_\mathrm{pp} p_\mathrm{sy\_pp\_ca_{out}} /
              \eta_\mathrm{pp\_ca}`

        '''

        def get_transmission(sy, nd, nd_2, ca, export=True):
            '''
            If called by supply rule, the order nd, nd_2 is always the ndcnn
            order, therefore also trm order.

            * **Case 1**: ``nd`` has higher time resolution (min) |rarr| just
              use ``trm[tm, sy, nd, nd_2, ca]``
            * **Case 2**: ``nd`` has lower time resolution (not min) |rarr|
              average ``avg(trm[tm, sy_2, nd, nd_2, ca])`` for all ``sy_2``
              defined by the
               ``grimsel.core.model_base.ModelBase.dict_sysy[nd, nd_2, sy]``

            Parameters
            ----------
            sy : int
                current time slot in nd
            nd : int
                outgoing node
            nd_2 : int
                incoming node
            ca : int
                energy carrier
            export : bool
                True if export else False

            '''
            if self.is_min_node[(nd if export else nd_2,
                                 nd_2 if export else nd)]:
                trm = self.trm[sy, nd, nd_2, ca]
                return trm

            else: # average over all of the other sy
                list_sy2 = self.dict_sysy[nd if export else nd_2,
                                          nd_2 if export else nd, sy]

                avg = 1/len(list_sy2) * sum(self.trm[_sy, nd, nd_2, ca]
                                            for _sy in list_sy2)
                return avg

        def supply_rule(self, sy, nd, ca):
            ''' Balance supply/demand '''

            list_neg = self.sll | self.curt
            prod = (# power output; negative if energy selling plant
                    sum(self.pwr[sy, pp, ca]
                        * (-1 if pp in list_neg else 1)
                        for (pp, nd, ca)
                        in set_to_list(self.ppall_ndca, [None, nd, ca]))
                    # incoming inter-node transmission
                    + sum(get_transmission(sy, nd, nd_2, ca, False)
                          / self.nd_weight[nd_2]
                          for (nd, nd_2, ca)
                          in set_to_list(self.ndcnn, [None, nd, ca]))
                   )
            exports = sum(get_transmission(sy, nd, nd_2, ca, True)
                          / self.nd_weight[nd]
                          for (nd, nd_2, ca)
                          in set_to_list(self.ndcnn, [nd, None, ca]))
            dmnd = (self.dmnd[sy, nd, ca]
                    + sum(self.pwr_st_ch[sy, st, ca] for (st, nd, ca)
                          in set_to_list(self.st_ndca, [None, nd, ca])))


            # demand of plants using ca as an input
            ca_cons = (po.ZeroConstant if not self.pp_ndcaca else
                       sum(self.pwr[sy, pp, ca_out] / self.pp_eff[pp, ca_out]
                           for (pp, nd, ca_out, ca)
                           in set_to_list(self.pp_ndcaca,
                                          [None, nd, None, ca])))
            gl = self.grid_losses[nd, ca]

            return prod == (dmnd + ca_cons) * (1 + gl) + exports

        self.cadd('supply', self.sy_ndca, rule=supply_rule)

    def add_energy_aggregation_rules(self):
        r'''
        Calculation of yearly totals from time slot power variables.

        * Total energy production by plant and output energy carrier:

          .. math::

             & E_\mathrm{p,c}
             = \sum_\mathrm{t} p_\mathrm{t,p,c} {w_\mathrm{\tau,t}} \\
             & \forall \mathrm{(p, c) \in ppall\_ca} \\

        * Total absolute ramping power:

        .. math::

           & |\Delta p_\mathrm{p,c}|
           = \sum_\mathrm{t} |\delta p_\mathrm{t,p,c}| \\
           & \forall \mathrm{(p, c) \in rp\_ca} \\

        * Total energy production by plant, fuel and output energy carrier:

          .. math::

             & E_\mathrm{p,n,c,f} = E_\mathrm{p,c} \\
             & \forall \mathrm{(p, n,c,f) \in pp\_ndcafl} \\

        '''

        def yearly_energy_rule(self, pp, ca):
            ''' Sets variable erg_yr for fuel-consuming plants. '''

            tm = self.dict_pp_tm_id[pp]

            return (self.erg_yr[pp, ca]
                    == sum(self.pwr[sy, pp, ca] * self.weight[tm, sy]
                           for tm, sy in set_to_list(self.tmsy, [tm, None])))

        self.cadd('yearly_energy', self.ppall_ca, rule=yearly_energy_rule)

        def yearly_ramp_rule(self, pp, ca):
            ''' Yearly ramping in MW/yrm, absolute aggregated up and down. '''

            tm = self.dict_pp_tm_id[pp]
            tmsy_list = set_to_list(self.tmsy, [tm, None])

            return (self.pwr_ramp_yr[pp, ca]
                    == sum(self.pwr_ramp_abs[sy, pp, ca]
                           for tm, sy in tmsy_list))

        self.cadd('yearly_ramping', self.rp_ca, rule=yearly_ramp_rule)

        def yearly_fuel_cons_rule(self, pp, nd, ca, fl):
            ''' Fuel consumed per plant and output energy carrier. '''

            return (self.erg_fl_yr[pp, nd, ca, fl]
                    == self.erg_yr[pp, ca])
        self.cadd('yearly_fuel_cons', self.pp_ndcafl,
                  rule=yearly_fuel_cons_rule)



    def add_capacity_calculation_rules(self):
        r'''
        Constraints concerning endogenous capacity calculations.

        * The total power capacity is composed of exogenous legacy capacity,
          optimized retirements, and optimized newly installed capacity:

          .. math::

             & P_\mathrm{tot, p, c} = P_\mathrm{leg, p, c}
             - P_\mathrm{ret, p, c} + P_\mathrm{new, p, c} \\
             & \forall \mathrm{(p,c)\in ppall\_ca}

        * The storage and reservoir energy capacity follows from the total
          power capacity and the fixed exogenous discharge duration.

          .. math::

             & C_\mathrm{tot, p,c} = P_\mathrm{tot, p,c} \zeta_\mathrm{p,c} \\
             & \forall \mathrm{(p,c) \in st\_ca \cup hyrs\_ca} \\

        '''



        def calc_cap_pwr_tot_rule(self, pp, ca):
            '''Calculate total power capacity (leg + add - rem).'''

            cap_tot = self.cap_pwr_leg[pp, ca]
            if pp in self.add:
                cap_tot += self.cap_pwr_new[pp, ca]
            if pp in self.rem:
                cap_tot -= self.cap_pwr_rem[pp, ca]
            return self.cap_pwr_tot[pp, ca] == cap_tot
        self.cadd('calc_cap_pwr_tot', self.ppall_ca,
                  rule=calc_cap_pwr_tot_rule)

        def calc_cap_erg_tot_rule(self, pp, ca):
            '''Calculate total energy capacity from total power capacity.'''

            return (self.cap_erg_tot[pp, ca]
                    == self.cap_pwr_tot[pp, ca]
                       * self.discharge_duration[pp, ca])
        self.cadd('calc_cap_erg_tot', self.st_ca | self.hyrs_ca,
                  rule=calc_cap_erg_tot_rule)

    def add_capacity_constraint_rules(self):
        r'''
        Power and stored energy are constrained by the respective capacities:

        * Output power of generators (note: variable renewables with
          exogenous supply (i.e. capacity factor) profile are explicitly
          excluded):

          .. math::

             & P_\mathrm{tot, p, c} \geqslant p_\mathrm{t, p, c} \\
             & \forall \mathrm{(t, p, c) \in
             (sy\_pp\_ca \setminus sy\_pr\_ca)
             \cup sy\_st\_ca \cup sy\_hyrs\_ca} \\

        * Charging power of storage assets:

          .. math::

             & P_\mathrm{tot, p, c} \geqslant p_\mathrm{chg,t,p,c} \\
             & \forall \mathrm{(t, p, c) \in sy\_st\_ca} \\

        * Stored energy of storage assets:

          .. math::

             & C_\mathrm{tot, p, c} \geqslant e_\mathrm{t,p,c} \\
             & \forall \mathrm{(t, p, c) \in sy\_st\_ca\cup sy\_hyrs\_ca} \\


        '''

        def ppst_capac_rule(self, sy, pp, ca):
            ''' Produced power must be less than capacity. '''

            if pp in self.setlst['pp'] and has_cap_avlb:

                tm = self.dict_pp_tm_id[pp]
                mt = self.dict_soy_month[(tm, sy)]

                cap_avlb = (self.cap_avlb[mt, pp, ca]
                            if self.dict_par['vc_fl'].has_monthly_factors
                            else self.cap_avlb[pp, ca])

                return (self.pwr[sy, pp, ca] <= self.cap_pwr_tot[pp, ca]
                                                * cap_avlb)
            else:
                return (self.pwr[sy, pp, ca] <= self.cap_pwr_tot[pp, ca])

        has_cap_avlb = hasattr(self, 'cap_avlb')
        self.cadd('ppst_capac',
                  (self.sy_pp_ca - self.sy_pr_ca) | self.sy_st_ca
                       | self.sy_hyrs_ca, rule=ppst_capac_rule)

        def st_chg_capac_rule(self, sy, pp, ca):
            ''' Charging power must be less than capacity. '''

            return (self.pwr_st_ch[sy, pp, ca]
                    <= self.cap_pwr_tot[pp, ca])

        self.cadd('st_chg_capac', self.sy_st_ca, rule=st_chg_capac_rule)

        def st_erg_capac_rule(self, sy, pp, ca):
            ''' Stored energy must be less than energy capacity. '''

            return (self.erg_st[sy, pp, ca]
                    <= self.cap_erg_tot[pp, ca])

        self.cadd('st_erg_capac', self.sy_st_ca | self.sy_hyrs_ca,
                  rule=st_erg_capac_rule)

        def pwr_pot_add_rule(self, pp, ca):
            ''' Capcity added + legacy must be less than potential capacity '''

            return (self.cap_pwr_tot[pp, ca]
                    <= self.pwr_pot[pp, ca])

        self.cadd('pwr_pot_add', self.add_ca, rule=pwr_pot_add_rule)

    def add_chp_rules(self):
        r'''
        Adds all co-generation related constraints.

        * Certain generators need to produce power following heat demand.
          This is implemented through the node-specific normalized CHP
          profile :math:`\phi_\mathrm{chp, sy\_pf}`: The production from
          power plants must be larger than the scaled CHP profile:

        .. math::

           & p_\mathrm{t,p,c} \geqslant \phi_\mathrm{chp,t,n,c}
           \mathrm{e}_\mathrm{chp,p,c} \\
           & \forall \mathrm{(t,p,c)} \in \mathrm{sy\_chp\_ca}


        '''

        def chp_prof_rule(model, sy, pp, ca):
            '''Produced power greater than CHP output profile.'''

            nd = self.mps.dict_plant_2_node_id[pp]

            return (self.pwr[sy, pp, ca]
                    >= self.chpprof[sy, nd, ca] * self.erg_chp[pp, ca])

        self.cadd('chp_prof', self.sy_chp_ca, rule=chp_prof_rule)


    def add_monthly_total_rules(self):
        r'''
        Adds the ``monthly_totals`` constraint which sets the monthly energy
        production variables:

        .. math::

           & E_\mathrm{m,p,c} = \sum_\mathrm{t\in sy_m} w_\mathrm{\tau(p),t}
           p_\mathrm{t,p,c} \\
           & \forall \mathrm{(m,p,c) \in mt \times hyrs\_ca} \\

        .. note::

           * :math:`\mathrm{sy_m}` is an ad-hoc set defining the time slots
             :math:`\mathrm{sy}` for any given month :math:`\mathrm{m}`.
           * The resulting monthly totals :math:`E_\mathrm{m,p,c}` are
             primarily used for hydro power constraints.

        '''


        def monthly_totals_rule(self, mt, pp, ca):
            '''Calculate monthly total production (hydro only). '''

            tm = self.dict_pp_tm_id[pp]

            list_sy = self.dict_month_soy[(tm, mt)]
            return (self.erg_mt[mt, pp, ca]
                    == sum(self.pwr[sy, pp, ca] * self.weight[tm, sy]
                           for sy in list_sy))

        if  len(set(self.df_tm_soy.mt_id)) == 12:
            self.cadd('monthly_totals', self.mt, self.hyrs_ca,
                      rule=monthly_totals_rule)
        else:
            logger.warning('Constraint monthly_totals: skipping. Temporal '
                           'model scope doesn\'t cover all months.')

    def add_variables_rules(self):
        r'''
        Produced power equals output profile.

        Variable renewable energy generators produce at a given output profile
        (capacity factor per time slot).

        .. note::
            Curtailments are included at the system level through dedicated
            technologies (set :math:`\mathrm{curt \subset ppall}`).

        .. math::

           & p_\mathrm{t,p,c} = \Phi_\mathrm{supply,t,p,c} P_\mathrm{tot,p,c}\\
           & \forall \mathrm{(t,p,c) \in sy\_pr\_ca}

        '''

        def variables_prof_rule(self, sy, pp, ca):
            ''' Produced power equal output profile '''
            left = self.pwr[sy, pp, ca]
            return left == (self.supprof[sy, pp, ca]
                            * self.cap_pwr_tot[pp, ca])

        self.cadd('variables_prof', self.sy_pr_ca, rule=variables_prof_rule)

    def add_ramp_rate_rules(self):
        r'''
        Three constraints are used to obtain the absolute hourly ramp rates
        :math:`|\delta_\mathrm{sy\_rp\_ca}|`:

        * Difference of power production (as in the
          :func:`add_charging_level_rules` constraint time is circular):

          .. math::

             & \delta_\mathrm{t,p,c} = p_\mathrm{t,p,c} - p_\mathrm{t - 1,p,c} \\
             & \forall \mathrm{(t,p,c) \in sy\_rp\_ca} \\

        * Two constraints to calculate the absolute values:

          .. math::

             & +1 \cdot \delta_\mathrm{t,p,c} \leqslant |\delta_\mathrm{t,p,c}| \\
             & -1 \cdot \delta_\mathrm{t,p,c} \leqslant |\delta_\mathrm{t,p,c}| \\
             & \forall \mathrm{(t,p,c) \in sy\_rp\_ca} \\

        '''


        def calc_ramp_rate_rule(self, sy, pp, ca):
            '''Ramp rates are power output differences.'''

            tm = self.dict_pp_tm_id[pp]

            list_sy = self.dict_tm_sy[tm]

            this_soy = sy
            last_soy = (sy - 1) if this_soy != list_sy[0] else list_sy[-1]

            return (self.pwr_ramp[sy, pp, ca]
                    == self.pwr[this_soy, pp, ca]
                     - self.pwr[last_soy, pp, ca])

        self.cadd('calc_ramp_rate', self.sy_rp_ca, rule=calc_ramp_rate_rule)


        def ramp_rate_abs_rule(self, sy, pp, ca):
            '''Standard LP absolute value constraints.'''

            return (flag_abs * self.pwr_ramp[sy, pp, ca]
                    <= self.pwr_ramp_abs[sy, pp, ca])

        flag_abs = 1
        self.cadd('ramp_rate_abs_pos', self.sy_rp_ca, rule=ramp_rate_abs_rule)
        flag_abs = -1
        self.cadd('ramp_rate_abs_neg', self.sy_rp_ca, rule=ramp_rate_abs_rule)

    def add_energy_constraint_rules(self):
        r'''
        Adds the ``pp_max_fuel`` constraint which limits the amount of
        output energy produced from certain fuels to the
        value of the exogenous parameter :math:`E_\mathrm{inp,n,c,f}`:

        .. math::

            & \sum_\mathrm{p \in ppall} E_\mathrm{p,n,c,f}
            \leqslant E_\mathrm{inp,n,c,f} \\
            & \forall \mathrm{(n,c,f) \in ndcafl} \\

        '''

        def pp_max_fuel_rule(self, nd, ca, fl):
            '''Constrain energy produced from certain fuels.'''

            is_constr = fl in self.fl_erg
            erg_inp_is_zero = self.erg_inp[nd, ca, fl] == 0

            ret = po.Constraint.Skip

            if is_constr and not erg_inp_is_zero:

                plant_list = set_to_list(self.ppall_ndcafl, [None, nd, ca, fl])

                if plant_list:

                    left = sum(self.erg_fl_yr[pp, nd_1, ca_1, fl_1]
                               for (pp, nd_1, ca_1, fl_1) in plant_list)
                    right = self.erg_inp[nd, ca, fl]
                    ret = left <= right

            return ret

        if hasattr(self, 'erg_inp'):
            self.cadd('pp_max_fuel', self.ndcafl, rule=pp_max_fuel_rule)

    def add_charging_level_rules(self):
        r'''
        Adds the constraint determining the stored energy: The energy in the
        current time slot :math:`e_\mathrm{t,p,c}` is equal what's left from
        the last time slot :math:`e_\mathrm{t-1,p,c}` plus

        * inflow minus production in the case of reservoirs and run-of-river
          plants
        * charging minus discharging in the case of pure storage plants
          without inflow

        Time is circular, i.e. the first time slot follows after the last.

        .. math::

           e_\mathrm{t,p,c} = e_\mathrm{t-t,p,c} +
           \begin{cases}
           (\phi_\mathrm{inflow,t,p} e_\mathrm{inp,n(p),c,f(p)} - p_\mathrm{t, p, c}) w_\mathrm{t, n(p)} \\
           \qquad \qquad \forall \mathrm{(t,p,c)\in sy\_hyrs\_ca \cup sy\_ror\_ca}\\
           (\eta_\mathrm{p,c}^{1/2} p_\mathrm{chg, t, p, c} - \eta_\mathrm{p,c}^{-1/2} p_\mathrm{t, p, c}) w_\mathrm{t} \\
           \qquad \qquad \forall \mathrm{(t,p,c)\in sy\_st\_ca}\\
           \end{cases}

        '''

        def erg_store_level_rule(self, sy, pp, ca):
            ''' Charging state for storage and hydro. '''

            nd = self.mps.dict_plant_2_node_id[pp]
            fl = self.mps.dict_plant_2_fuel_id[pp]
            tm = self.dict_nd_tm_id[nd]

            list_sy = self.dict_tm_sy[tm]

            this_soy = sy
            last_soy = (sy - 1) if this_soy != list_sy[0] else list_sy[-1]

            left = 0
            right = 0

            # last time slot's energy level for storage and hyrs
            # this excludes run-of-river, which doesn't have an energy variable
            if pp in self.setlst['st'] + self.setlst['hyrs']:
                left += self.erg_st[this_soy, pp, ca] # in MWh of stored energy
                right += self.erg_st[last_soy, pp, ca] #* (1-self.st_lss_hr[pp, ca])

            if pp in self.setlst['st']:
                right += ((- self.pwr[this_soy, pp, ca]
                           / (1 - self.st_lss_rt[pp, ca])**(1/2)
                           * self.weight[tm, sy]
                         ) + (
                           self.pwr_st_ch[this_soy, pp, ca]
                           * (1 - self.st_lss_rt[pp, ca])**(1/2)
                           * self.weight[tm, sy]))
            elif pp in self.setlst['hyrs'] + self.setlst['ror']:
                right += (
                          # inflowprof profiles are normalized to one!!
                          (self.inflowprof[this_soy, pp, ca]
                           * self.erg_inp[nd, ca, fl]
                          - self.pwr[sy, pp, ca]) * self.weight[tm, sy]
                         )
            return left == right

        self.cadd('erg_store_level',
                  self.sy_st_ca | self.sy_hyrs_ca | self.sy_ror_ca,
                  rule=erg_store_level_rule)


    def add_hydro_rules(self):
        r'''
        Various rules constraining the operation of the hydro reservoir plants.

        * The ``hy_reservoir_boundary_conditions`` constraint requires
          the reservoir filling level to assume an exogenously defined share
          of the energy capacity during certain time slots:

          .. math::

             & \rho_\mathrm{t,p,c}
             = e_\mathrm{hyd\_bc,p,c} C_\mathrm{tot, p, c} \\
             & \forall \mathrm{(t,p,c) \in sy\_hydbc \times hyrs\_ca} \\

        * The ``hy_month_min`` constraint forces a certain minimum output
          production from hydro reservoirs each month. This minimum is
          expressed as a share of maximum monthly inflow (which itself is
          a share :math:`\rho_\mathrm{max\_erg\_in,p}` of the total yearly
          inflow energy :math:`e_\mathrm{inp,n,c,f}`):

          .. math::

             E_\mathrm{m, p, c} \geqslant & \rho_\mathrm{max\_erg\_in,p}\\
                                & \cdot \rho_\mathrm{min\_erg\_out,p}\\
                                & \cdot e_\mathrm{inp,n(p),c,f(p)}\\

        * The ``hy_erg_min`` constraint sets a lower bound on the stored
          energy as a fraction of the total energy capacity.

          .. math::

             e_\mathrm{t,p,c} \geqslant & C_\mathrm{tot,p,c} \\
                                        & \cdot \rho_\mathrm{min\_cap,p} \\
                                        & \forall \mathrm{(t,p,c)
                                        \in sy\_hyrs\_ca}

        '''


        def hy_reservoir_boundary_conditions_rule(self, sy, pp, ca):
            '''Reservoirs stored energy boundary conditions.'''

            if (sy, pp) in [i for i in self.hyd_erg_bc.sparse_iterkeys()]:
                return (self.erg_st[sy, pp, ca]
                        == self.hyd_erg_bc[sy, pp]
                           * self.cap_erg_tot[pp, ca])
            else:
                return po.Constraint.Skip

        if hasattr(self, 'hyd_erg_bc'):
            self.cadd('hy_reservoir_boundary_conditions', self.sy_hyrs_ca,
                      rule=hy_reservoir_boundary_conditions_rule)

        def hy_month_min_rule(self, mt, pp, nd, ca, fl):
            '''Reservoirs minimum monthlyl power production.'''

            return (self.erg_mt[mt, pp, ca]
                    >= self.max_erg_mt_in_share[pp]
                     * self.min_erg_mt_out_share[pp]
                     * self.erg_inp[nd, ca, fl])

        self.cadd('hy_month_min', self.mt, self.hyrs_ndcafl,
                  rule=hy_month_min_rule)

        def hy_erg_min_rule(self, sy, pp, ca):
            '''Reservoirs minimum filling level.'''

            if not pp in [h for h in self.min_erg_share]:
                return po.Constraint.Skip
            else:
                return (self.erg_st[sy, pp, ca]
                        >=
                        self.min_erg_share[pp]
                        * self.cap_erg_tot[pp, ca])

        self.cadd('hy_erg_min', self.sy_hyrs_ca, rule=hy_erg_min_rule)

    def add_yearly_cost_rules(self):
        r'''
        Groups the yearly cost calculation constraints:

        * Variable fuel costs :math:`c_\mathrm{fuel,p,c,f}` of
          plants with constant supply curves:

          - For fixed fuel costs :math:`\mathrm{vc}_\mathrm{fuel,fl\_nd}`:

             .. math::

                & c_\mathrm{fuel,p,c,f}
                = E_\mathrm{p,n,c,f} \mathrm{vc}_\mathrm{f,n} \\
                & \forall \mathrm{(p,n,c,f)\in ppall\_ndcafl
                                            \setminus lin\_ndcafl} \\

          - For monthly fuel costs :math:`vc_\mathrm{m,d,n}`:

             .. math::

                & c_\mathrm{fuel,p,c,f}
                = \sum_\mathrm{t\in sy\_pp\_ca|_{p,c}} \mathrm{vc_{m(t),f,n}}
                  p_\mathrm{t,p,c} / \eta_\mathrm{p,c}\\
                & \forall \mathrm{(p,n,c,f)\in ppall\_ndcafl
                                             \setminus lin\_ndcafl} \\

          - For plants with cost profiles:

              .. math::

                & c_\mathrm{fuel,p,c,f}
                = \sum_\mathrm{t\in sy\_ppall\_ca|_{p,c}}
                p_\mathrm{t,p,c} / \eta_\mathrm{p,c}
                w_\mathrm{\tau(p),t} \cdot
                \begin{cases}
                -1 \cdot \Phi_\mathrm{psll,t,\phi(p)}
                    & \text{if } \mathrm{p\in sll} \\
                +1 \cdot \Phi_\mathrm{pbuy,t,\phi(p)}
                    & \text{if } \mathrm{p\notin sll} \\
                \end{cases} \\
                & \forall \mathrm{(p,n,c,f)
                \in ppall\_ndcafl \setminus lin\_ndcafl} \\

        * Variable operation and maintenance costs
          :math:`\mathrm{vc}_\mathrm{om,p,c}` for all plants:

             .. math::

                & c_\mathrm{om_v,p,c}
                = E_\mathrm{p,c} \mathrm{vc}_\mathrm{om,p,c} \\
                & \forall \mathrm{(p,c)\in ppall\_ca} \\

        * Variable |CO2|  emission costs: Same as variable fuel costs
          (fixed fuel costs and monthly fuel costs cases), but with
          specific cost
           :math:`\pi_\mathrm{CO_2,[m],n}\cdot i_\mathrm{CO_2, f}`.

        * Total ramping cost:

          .. math::

             & c_\mathrm{rp,p,c}
             = |\Delta_\mathrm{p,c}| \mathrm{vc}_\mathrm{ramp,p,c} \\
             & \forall \mathrm{(p,c)\in rp\_ca} \\

        * Total fixed O&M cost (only relevant for plants with capacity
          additions and retirements):

          .. math::

             & c_\mathrm{om_f,p,c}
             = P_\mathrm{tot,p,c} \mathrm{fc}_\mathrm{om,p,c} \\
             & \forall \mathrm{(p,c)\in add\_ca \cup rem\_ca} \\


        * Total annualized fixed investment cost:

          .. math::

             & c_\mathrm{cp,p,c}
             = P_\mathrm{new,p,c} {c}_\mathrm{cp,p,c} \\
             & \forall \mathrm{(p,c)\in add\_ca} \\


        .. note::
           The fuel and emission costs of power plants with linear supply
           curves must be calculated directly in the objective function.
           This is because CPLEX only supports a quadratic objective, not
           quadratic constraints.


        '''


        def calc_vc_fl_pp_rule(self, pp, nd, ca, fl):
            '''Yearly fuel cost calculation (constant supply curve plants).'''

            tm = self.dict_nd_tm_id[nd]
            list_sy = self.dict_tm_sy[tm]

            sign = -1 if pp in self.sll else 1

            # Case 1: fuel has price profile
            if (fl, nd, ca) in {**self.dict_pricebuy_pf,
                                **self.dict_pricesll_pf}:

                pprf = (self.pricesllprof if sign is -1 else self.pricebuyprof)

                pf = (self.dict_pricesll_pf[(fl, nd, ca)] if sign is -1 else
                      self.dict_pricebuy_pf[(fl, nd, ca)])

                sums = (sign * sum(self.weight[tm, sy]
                                   * pprf[sy, pf]
                                   / self.pp_eff[pp, ca]
                                   * self.pwr[sy, pp, ca]
                                   for (sy, _, _)
                                   in set_to_list(self.sy_pp_ca,
                                                  [None, pp, ca])))

            # Case 2: monthly adjustment factors have been applied to vc_fl
            elif self.dict_par['vc_fl'].has_monthly_factors:
                sums = (sign * sum(self.weight[tm, sy]
                                   * self.vc_fl[self.dict_soy_month[(tm, sy)], fl, nd]
                                   / self.pp_eff[pp, ca]
                                   * self.pwr[sy, pp, ca] for (sy, _pp, _ca)
                                   in set_to_list(self.sy_pp_ca,
                                                  [None, pp, ca])))

            # Case 3: ordinary single fuel price
            else:
                sums = (sign * self.erg_fl_yr[pp, nd, ca, fl]
                             / self.pp_eff[pp, ca]
                             * self.vc_fl[fl, nd])

            return self.vc_fl_pp_yr[pp, ca, fl] == sums

        self.cadd('calc_vc_fl_pp', self.pp_ndcafl - self.lin_ndcafl,
                  rule=calc_vc_fl_pp_rule)

        def calc_vc_om_pp_rule(self, pp, ca):
            '''Yearly variable O&M cost calculation rule.'''

            return (self.erg_yr[pp, ca] * self.vc_om[pp, ca]
                    == self.vc_om_pp_yr[pp, ca])

        if hasattr(self, 'vc_om'):
            self.cadd('calc_vc_om_pp', self.ppall_ca, rule=calc_vc_om_pp_rule)


        def calc_vc_co2_pp_rule(self, pp, nd, ca, fl):
            '''Yearly emission ncost calculation (constant supply curves).'''

            tm = self.dict_nd_tm_id[nd]

            # Case 1: monthly adjustment factors have been applied to vc_fl
            if self.dict_par['price_co2'].has_monthly_factors:
                sums = sum(self.pwr[sy, pp, ca] # POWER!
                           / self.pp_eff[pp, ca] * self.weight[tm, sy]
                           * self.price_co2[mt, nd] * self.co2_int[fl]
                           for (_tm, sy, mt) in set_to_list(self.tmsy_mt,
                                                       [tm, None, None]))
            # Case 2: ordinary single CO2 price
            else:
                sums = (self.erg_fl_yr[pp, nd, ca, fl] # ENERGY!
                            / self.pp_eff[pp, ca]
                            * self.price_co2[nd] * self.co2_int[fl])

            return self.vc_co2_pp_yr[pp, ca] == sums

        self.cadd('calc_vc_co2_pp', self.pp_ndcafl - self.lin_ndcafl,
                  rule=calc_vc_co2_pp_rule)

        def calc_vc_ramp_rule(self, pp, ca):
            '''Yearly ramping variable cost calculation rule.'''

            return (self.vc_ramp_yr[pp, ca]
                    == self.pwr_ramp_yr[pp, ca] * self.vc_ramp[pp, ca])

        if hasattr(self, 'vc_ramp'):
            self.cadd('calc_vc_ramp', self.rp_ca, rule=calc_vc_ramp_rule)

        def calc_fc_om_rule(self, pp, ca):
            '''Fixed O&M cost calculation rule.'''

            return (self.fc_om_pp_yr[pp, ca]
                    == self.cap_pwr_tot[pp, ca] * self.fc_om[pp, ca])

        if hasattr(self, 'fc_om'):
            self.cadd('calc_fc_om', self.add_ca | self.rem_ca,
                      rule=calc_fc_om_rule)

        def calc_fc_cp_rule(self, pp, ca):
            '''Fixed capital cost calculation rule'''

            return (self.fc_cp_pp_yr[pp, ca]
                    == self.cap_pwr_new[pp, ca] * self.fc_cp_ann[pp, ca])

        self.cadd('calc_fc_cp', self.add_ca, rule=calc_fc_cp_rule)

    def add_objective_rules(self):
        ''' Quadratic objective function.


        The objective function to be minimized is the sum of all system
        costs:

        * Fuel costs of power plants with constant efficiency/supply curve:

        .. math::
            \sum_\mathrm{(p,c,f) \in pp\_cafl \setminus lin\_cafl}
                    c_\mathrm{fuel,p,c,f}

        * |CO2| emission costs of power plants with constant efficiency:

        .. math::
            \sum_\mathrm{(p,c) \in pp\_ca \setminus lin\_ca} c_\mathrm{em,p,c}

        * Variable O\&M costs of power plants:

        .. math::
            \sum_\mathrm{(p,c) \in ppall\_ca} c_\mathrm{om_v,p,c}

        * Fixed O\&M costs of power plants:

        .. math::
            \sum_\mathrm{(p,c) \in ppall\_ca} c_\mathrm{om_f,p,c}


        * Variable ramping costs of power plants:

        .. math::
            \sum_\mathrm{(p,c) \in rp\_ca} c_\mathrm{rp,p,c}

        * Fixed investment costs of power plants:

          .. math::
             \sum_\mathrm{(p,c) \in add\_ca} c_\mathrm{cp,p,c}

        * Variable fuel costs for plants with linear supply curves are
          calculated in the method :func:`get_vc_fl`.

        * Variable |CO2| emission costs for plants with linear supply curves
          are calculated in the method :func:`get_vc_co`.

        .. note::
           The variable fuel and emission cost terms of the power plants with
           linear supply curves are not model variables but calculated directly
           in the auxiliary methods
           :func:`get_vc_fl` and :func:`get_vc_fl`. See the note in the
           :func:`add_yearly_cost_rules` method documentation.

        '''

        def objective_rule_quad(self):

            return (# FUEL COST CONSTANT
                    sum(self.vc_fl_pp_yr[pp, ca, fl]
                        * self.nd_weight[self.mps.dict_plant_2_node_id[pp]]
                        for (pp, ca, fl)
                        in set_to_list(self.pp_cafl - self.lin_cafl, nnn))
                    # FUEL COST LINEAR
                  + self.get_vc_fl()
                    # EMISSION COST LINEAR
                  + self.get_vc_co()
                  + sum(self.vc_co2_pp_yr[pp, ca]
                        * self.nd_weight[self.mps.dict_plant_2_node_id[pp]]
                        for (pp, ca) in set_to_list(self.pp_ca - self.lin_ca, nn))
                  + sum(self.vc_om_pp_yr[pp, ca]
                        * self.nd_weight[self.mps.dict_plant_2_node_id[pp]]
                        for (pp, ca) in set_to_list(self.ppall_ca, nn))
                  + sum(self.vc_ramp_yr[pp, ca]
                        * self.nd_weight[self.mps.dict_plant_2_node_id[pp]]
                        for (pp, ca) in set_to_list(self.rp_ca, nn))
                  + sum(self.fc_om_pp_yr[pp, ca]
                        * self.nd_weight[self.mps.dict_plant_2_node_id[pp]]
                        for (pp, ca) in set_to_list(self.ppall_ca, nn))
                  + sum(self.fc_cp_pp_yr[pp, ca]
                        * self.nd_weight[self.mps.dict_plant_2_node_id[pp]]
                        for (pp, ca) in set_to_list(self.add_ca, nn)))

        self.cadd('objective_quad', rule=objective_rule_quad,
                  sense=po.minimize, objclass=po.Objective)

# %%
#
#plin = ml.io.modwr.dict_comp_obj['pwr'].get_df().rename(columns={'value': 'pwr'})#.query('pp_id in %s'%ml.m.setlst['lin'])
#
#plin['fl_id'] = plin.pp_id.replace(ml.m.mps.dict_plant_2_fuel_id)
#plin['nd_id'] = plin.pp_id.replace(ml.m.mps.dict_plant_2_node_id)
#plin['tm_id'] = plin.nd_id.replace(ml.m.dict_nd_tm_id)
#
#plin = plin.join(ml.m.df_def_plant.set_index(['pp_id'])[['set_def_lin', 'set_def_sll']], on=['pp_id'])
#
#plin = plin.join(ml.m.df_tm_soy.set_index(['tm_id', 'sy'])[['mt_id', 'weight']], on=['tm_id', 'sy'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['vc_fl'].get_df().set_index(['mt_id', 'fl_id', 'nd_id']).value.rename('vc_fl'), on=['mt_id', 'fl_id', 'nd_id'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['vc_om'].get_df().set_index(['pp_id']).value.rename('vc_om'), on=['pp_id'])
#
#plin = plin.join(ml.io.modwr.dict_comp_obj['factor_lin_0'].get_df().set_index(['pp_id']).value.rename('f0'), on=['pp_id'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['factor_lin_1'].get_df().set_index(['pp_id']).value.rename('f1'), on=['pp_id'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['nd_weight'].get_df().set_index(['nd_id']).value.rename('nd_weight'), on=['nd_id'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['pp_eff'].get_df().set_index(['pp_id']).value.rename('pp_eff'), on=['pp_id'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['co2_int'].get_df().set_index(['fl_id']).value.rename('co2_int'), on=['fl_id'])
#plin = plin.join(ml.io.modwr.dict_comp_obj['price_co2'].get_df().set_index(['nd_id', 'mt_id']).value.rename('price_co2'), on=['nd_id', 'mt_id'])
#
#prfsll = ml.io.modwr.dict_comp_obj['pricesllprof'].get_df()
#prfsll['fl_id'] = prfsll.pf_id.replace({v: k[0] for k, v in ml.m.dict_pricesll_pf.items()})
#prfbuy = ml.io.modwr.dict_comp_obj['pricebuyprof'].get_df()
#prfbuy['fl_id'] = prfbuy.pf_id.replace({v: k[0] for k, v in ml.m.dict_pricebuy_pf.items()})
#
#plin = plin.join(prfsll.set_index(['sy', 'fl_id']).value.rename('prfsll'), on=['sy', 'fl_id'])
#plin = plin.join(prfbuy.set_index(['sy', 'fl_id']).value.rename('prfbuy'), on=['sy', 'fl_id'])
#
#plin = ml.m.mps.id_to_name(plin)
#
#plin['vcom'] = plin.eval('pwr * weight * vc_om * nd_weight')
#
#plin['c_trm'] = 1e4
#plin['c_trm']  = plin.c_trm.where(plin.fl == 'electricity', 0)
#plin['CC_trm'] = (plin.pwr**2) * plin.nd_weight * plin.c_trm
#
#plin.loc[plin.fl == 'electricity']
#
##plin.assign(pp_id=lambda x: x.pp_id.astype(np.int)).groupby('pp_id').sum().reset_index().join(dfan.set_index('pp_id'), on='pp_id').set_index('pp_id').iloc[:,-2:].reset_index().assign(dff=lambda x: x.diff(axis=1).iloc[:, -1], pp=lambda x: x.pp_id.replace(ml.m.mps.dict_pp)).sort_values('dff').dff
#
#VC_FL_L = plin.query('set_def_lin == 1').eval('pwr * weight * vc_fl * nd_weight * (f0 + 0.5 * pwr * f1)').sum()
#VC_FL_C = plin.query('set_def_lin == 0').eval('pwr * weight * vc_fl * nd_weight / pp_eff').sum()
#VC_CO_L = plin.query('set_def_lin == 1').eval('pwr * weight * co2_int * price_co2 * nd_weight * (f0 + 0.5 * pwr * f1)').sum()
#VC_CO_C = plin.query('set_def_lin == 0').eval('pwr * weight * co2_int * price_co2 * nd_weight / pp_eff').sum()
#VC_OM = plin.eval('pwr * weight * vc_om * nd_weight').sum()
#VC_RAMP = ml.io.modwr.dict_comp_obj['vc_ramp_yr'].get_df().value.sum()
#VC_TRM = plin.query('fl == "electricity"').CC_trm.sum()
#
#VC_SLL_HH = - plin.query('fl == "electricity" and set_def_sll == 1').eval('nd_weight * weight * pwr * prfsll').sum()
#VC_BUY_HH = plin.query('fl == "electricity" and set_def_sll == 0').eval('nd_weight * weight * pwr * prfbuy').sum()
#
#VC_TRM + (VC_FL_L + VC_FL_C + VC_CO_L + VC_CO_C + VC_OM + VC_RAMP + VC_SLL_HH + VC_BUY_HH)
#
#ml.m.objective_value

# %%


    def get_vc_fl(self):
        r'''
        Get total fuel cost calculated directly from power production:

        .. math::
           \sum_\mathrm{(t,p,c)\in sy\_lin\_ca}
           p_\mathrm{t,p,c} w_\mathrm{\tau(p),t}
           \cdot \mathrm{vc_{f(p),n(p)}}
           \cdot (f_\mathrm{0,p,c} + 0.5 p_\mathrm{t,p,c} f_\mathrm{1,p,c})

        '''


        def spec_vc_fl(lin, sy):

            fl_id, nd_id = (self.mps.dict_plant_2_fuel_id[lin],
                            self.mps.dict_plant_2_node_id[lin])

            if self.dict_par['vc_fl'].has_monthly_factors:
                mt_id = self.dict_soy_month[(self.dict_pp_tm_id[lin], sy)]
                return self.vc_fl[mt_id, fl_id, nd_id]
            else:
                return self.vc_fl[fl_id, nd_id]

        return \
        sum(self.pwr[sy, lin, ca]
            * self.weight[self.dict_pp_tm_id[lin], sy]
            * spec_vc_fl(lin, sy)
            * (self.factor_lin_0[lin, ca]
               + 0.5 * self.pwr[sy, lin, ca]
                     * self.factor_lin_1[lin, ca])
            * self.nd_weight[self.mps.dict_plant_2_node_id[lin]]
            for (sy, lin, ca) in set_to_list(self.sy_lin_ca, nnn))

    def get_vc_co(self):
        r'''
        Get total |CO2| emission cost calculated directly from power
        production:

        .. math::
           \sum_\mathrm{(t,p,c)\in sy\_lin\_ca}
           p_\mathrm{t,p,c} w_\mathrm{\tau(p),t}
           \cdot \pi_\mathrm{CO_2, m(t), n(p)} i_\mathrm{CO_2,f}
           \cdot (f_\mathrm{0,p,c} + 0.5 p_\mathrm{t,p,c} f_\mathrm{1,p,c})

        '''

        return \
        sum(self.pwr[sy, lin, ca] * self.weight[self.dict_pp_tm_id[lin], sy]
            * (self.price_co2[self.dict_soy_month[(self.dict_pp_tm_id[lin], sy)],
                              self.mps.dict_plant_2_node_id[lin]]
               if self.dict_par['price_co2'].has_monthly_factors
               else self.price_co2[self.mps.dict_plant_2_node_id[lin]])
            * self.co2_int[self.mps.dict_plant_2_fuel_id[lin]]
                * (self.factor_lin_0[lin, ca]
                   + 0.5 * self.pwr[sy, lin, ca] * self.factor_lin_1[lin, ca])
            * self.nd_weight[self.mps.dict_plant_2_node_id[lin]]
            for (sy, lin, ca) in set_to_list(self.sy_lin_ca, nnn))





import itertools
import pandas as pd

def pdef(df, sets=[], val='value', fixnan=0):
    '''
    Transform a dataframe into a dictionary;
    primarily used for parameter setting.

    Keyword arguments:
    df -- input dataframe
    sets -- selection of columns (default all columns except val)
    val -- value column (default "value")
    fixnan -- If not False: used as keyword argument for fillna;
              if False: drop rows with null values
    '''
    df = df.reset_index(drop=True)
    sets = [c for c in df.columns if not c == val] if len(sets) == 0 else sets
    df = df[sets + [val]]

    if not (type(fixnan) == bool and fixnan == False):
        df = df.fillna(fixnan)
    else:
        df = df.loc[-df[val].isnull()]
    dct = df.set_index(sets)[val].to_dict()
    return dct

def set_to_list(st, vl):
    '''
    Takes a (multi-dimensional) pyomo set and returns all index combinations
    which satisfy the conditions vl
    Input:
    - st: pyomo set, e.g. pp_ndca composed of pp, nd, ca
    - vl: tuple with same dimensions, e.g. (None, None, 0) for all set members with ca == 0
    '''
    vl = [[v] if not (type(v) == list or v == None)  else v for v in vl]
    ind = [i[0] for i in enumerate(vl) if not i[1]==None]
    cbs = list(itertools.product(*[v for v in vl if not v == None]))
    return [a for a in st if tuple([a[ia] for ia in ind]) in cbs]

def cols2tuplelist(*args, return_df=False):
    '''
    Converts dataframes to lists of tuples.

    Multiple inputs are joined by crossproducts, i.e. yielding all
    combinations of the input rows.
    '''

    tl = []
    cols = []
    for idf in args:

        if type(idf) == pd.core.frame.Series:
            idf = pd.DataFrame(idf)

#        appkwargs = dict(func=tuple, axis=1)

        cols += idf.columns.tolist()

#        tl.append(list(idf.drop_duplicates().apply(**appkwargs)))
        tl.append([tuple(row) for row in idf.drop_duplicates().values])

    prod = list(itertools.product(*tl))
    prod = [tuple([ccc for cc in c for ccc in cc]) for c in prod]

    if return_df:
        prod = pd.DataFrame(prod, columns=cols)
    return prod



def get_ilst(df, cols=None):
    '''
    Get index list from dataframe columns, using all or a selection of the
    columns.
    - df: input dataframe
    - cols: selection of columns
    '''
#    cols = df.columns if cols == None else cols
#    return [tuple(l) for l in
#            df.loc[:, cols].drop_duplicates().get_values()]
    print('get_ilst got superseeded by cols2tuplelist')
    return cols2tuplelist(df[cols] if cols else df)

